"""
Manage interactions with LLMs across providers.
Provides functional configuration, initialization, 
and response processing/retry/fallback mechanisms for making LLM calls
through a unified interface that works with OpenAI, Anthropic, Amazon Bedrock, Google, and others.
"""
from functools import wraps

from typing import Any
from .llm_manager import LLMManager

@wraps(LLMManager.configure_llm_backend)
def configure_llm_backend(
    provider: str,
    model: str,
    **kwargs
) -> None:
    """Wrapper for LLMManager.configure_llm_backend"""
    LLMManager.configure_llm_backend(provider, model, **kwargs)


@wraps(LLMManager.get_llm_config)
def get_llm_config() -> dict[str, Any]:
    """Wrapper for LLMManager.get_llm_config"""
    return LLMManager.get_llm_config()


@wraps(LLMManager.call_llm)
def call_llm(messages: list[dict[str, str]], **kwargs) -> str:
    """Wrapper for LLMManager.call_llm"""
    return LLMManager.call_llm(messages, **kwargs)


@wraps(LLMManager.retry_call_llm)
def retry_call_llm(
    messages: list[dict[str, str]],
    process_response,
    failure_handler,
    max_attempts: int = 5,
    call_llm_kwargs: dict[str, Any] | None = None,
    process_response_kwargs: dict[str, Any] | None = None,
    failure_handler_kwargs: dict[str, Any] | None = None,
) -> Any:
    """Wrapper for LLMManager.retry_call_llm"""
    return LLMManager.retry_call_llm(
        messages=messages,
        process_response=process_response,
        failure_handler=failure_handler,
        max_attempts=max_attempts,
        call_llm_kwargs=call_llm_kwargs,
        process_response_kwargs=process_response_kwargs,
        failure_handler_kwargs=failure_handler_kwargs,
    )



"""
This module contains functions to generate and annotate gene lists.
"""

from .make_cell_type_gene_list import (
    ai_make_cell_type_gene_list,

)

__all__ = [
    #make_cell_type_gene_list.py
    "ai_make_cell_type_gene_list"
]



"""
This module contains functions that calculate cell type marker gene scores automaticallly (i.e. you supply only the cell type, not the marker genes).
"""
import numpy as np
import scanpy as sc

from anndata import AnnData

from anndict.utils.anndata_ import filter_gene_list
from anndict.annotate.genes import ai_make_cell_type_gene_list


def cell_type_marker_gene_score(
    adata: AnnData,
    cell_type_col: str = None,
    cell_types: list[str] = None,
    species: str = "Human",
    list_length: str = None,
    score_name: str = "_score",
    adt_key: tuple = None,
    **kwargs,
) -> None:
    """
    Compute marker gene scores for specified cell types. Must provide 
    either a list of cell types, or a column that contains cell_type labels.

    Parameters
    ------------
    adata
        An :class:`AnnData` object.

    cell_type_col
        Column name in ``adata.obs`` containing cell type annotations.

    cell_types
        List of cell types for which to compute the marker gene scores.

    species
        Species for gene list generation. Defaults to ``'Human'``.

    list_length
        Qualitative length of the marker gene list. Can be anything like 
        ``'short'`` or ``'long'``. Try ``'long'`` if you are having trouble 
        getting valid genes that are present in your dataset.

    score_name
        Suffix for the computed score names. Defaults to '_score'.

    adt_key
        Used by :func:`adata_dict_fapply` or :func:`adata_dict_fapply_return` 
        when passing this function to them.

    **kwargs
        Optional keyword args passed to :func:`sc.tl.score_genes`.

    Returns
    --------
    This function modifies the input ``adata`` object in-place. See **Notes** for details.

    Notes
    ------
        - ``adata.var`` is updated with boolean columns indicating genes used in the scores.
        - ``adata.obs`` is updated with new columns containing the computed scores for each 
            observation.

    See Also
    ---------
    :func:`module_score_umap` : To conveniently visualize these module scores on UMAPs.
    :func:`module_score_barplot` : To conveniently visualize these module scores as a bar plot.
    """

    score_name_suffix = score_name

    # Check for conflicting parameters
    if cell_types is not None and cell_type_col is not None:
        raise ValueError("Provide either 'cell_type_col' or 'cell_types', not both.")

    if cell_types is None:
        if cell_type_col is not None:
            cell_types = adata.obs[cell_type_col].unique().tolist()
        else:
            raise ValueError("Either 'cell_type_col' or 'cell_types' must be provided.")
    else:
        # Ensure cell_types is a list
        if isinstance(cell_types, str):
            cell_types = [cell_types]

    for cell_type in cell_types:
        cell_type = str(cell_type)  # Ensure cell_type is a string
        # Set the score_name per cell type
        score_name = f"{cell_type}{score_name_suffix}"

        # Generate gene list using ai_gene_list function
        gene_list = ai_make_cell_type_gene_list(cell_type, species, list_length=list_length)

        # Filter the gene list based on genes present in adata
        gene_list = filter_gene_list(adata, gene_list)

        # Mark genes included in this score in adata.var
        adata.var[score_name] = adata.var.index.isin(gene_list)

        # calculate score if any valid genes, otherwise print warning and assign score value as NaN.
        if gene_list:
            # Compute the gene score and store it in adata.obs[score_name]
            sc.tl.score_genes(adata, gene_list=gene_list, score_name=score_name, **kwargs)
        else:
            # Assign NaN to adata.obs[score_name] for all observations
            adata.obs[score_name] = np.nan
            print(f"No valid genes for {cell_type} in {adt_key if adt_key else ''}. Assigning score value as NaN")



"""
This module contains utility functions for ``AnnDictionary``.
"""

from .anndictionary_ import (
    enforce_semantic_list,
    make_names,
    normalize_string,
    normalize_label,
    create_color_map,
    get_slurm_cores,
    summarize_metadata,
    display_html_summary,

)

from .anndata_ import (
    remove_genes,
    add_col_to_adata_obs,
    add_col_to_adata_var,
    convert_obs_col_to_category,
    convert_obs_col_to_string,
    convert_obs_index_to_str,
    get_adata_columns,

)

from .pca_density_filter import (
    pca_density_filter_main,
    pca_density_filter_adata,
    pca_density_subsets,

)

from .read_spatial_data import (
    read_transcript_coords,
    get_steps_and_coords,
    populate_sparse_array,
    process_gene_counts,
    create_anndata,
    add_blank_image_to_adata,
    build_adata_from_transcript_positions,
    build_adata_from_visium,
)

from .uce_ import (
    uce_adata,

)

__all__ = [
    #anndictionary_.py
    "enforce_semantic_list",
    "make_names",
    "normalize_string",
    "normalize_label",
    "create_color_map",
    "get_slurm_cores",
    "summarize_metadata",
    "display_html_summary",

    # anndata_.py
    "remove_genes",
    "add_col_to_adata_obs",
    "add_col_to_adata_var",
    "convert_obs_col_to_category",
    "convert_obs_col_to_string",
    "convert_obs_index_to_str",
    "get_adata_columns",

    # pca_density_filter.py
    "pca_density_filter_main",
    "pca_density_filter_adata",
    "pca_density_subsets",

    # read_spatial_data.py
    "read_transcript_coords",
    "get_steps_and_coords",
    "populate_sparse_array",
    "process_gene_counts",
    "create_anndata",
    "add_blank_image_to_adata",
    "build_adata_from_transcript_positions",
    "build_adata_from_visium",

    # uce_.py
    "uce_adata",
]


"""
This subpackage contains functions to compare labels.
"""

#label comparison utility functions
from .label_utils import (
   create_label_df,
)

#functions to calculate agreement between label columns
from .agreement_metrics import (
    kappa_adata,
    krippendorff_alpha_adata,
)

#ai methods to compare labels
from .ai_compare_cell_type_labels import (
   ai_compare_cell_type_labels_pairwise,
)


__all__ = [
    # label_utils.py
    "create_label_df",

    # agreement_metrics.py
    "kappa_adata",
    "krippendorff_alpha_adata",

    # ai_compare_cell_type_labels.py
    "ai_compare_cell_type_labels_pairwise",
]


"""
This module contains core functions for de novo annotation of cells based on marker genes and LLMs.
The functions in this module are called by other annotation functions.
We include these functions in the docs for reference, but you should not generally use them directly.
"""
import warnings
import pandas as pd
import scanpy as sc

from pandas import DataFrame
from anndata import AnnData

from anndict.utils.anndata_ import convert_obs_col_to_category

def _validate_and_prepare_rank_genes(
    adata: AnnData,
    groupby: str
) -> tuple:
    """
    Validates the ``groupby`` column and ensures ``rank_genes_groups`` analysis is present.

    Parameters
    ----------
    adata
        An :class:`AnnData`.

    groupby
        The column name to group by
        
    Returns
    -------
    A :class:`tuple` of the form ``(rank_genes_groups, clusters)``
    """
    # Ensure the groupby column is categorical
    if not isinstance(adata.obs[groupby].dtype, pd.CategoricalDtype):
        adata.obs[groupby] = adata.obs[groupby].astype('category')

    # Get the number of categories in the groupby column
    n_categories = len(adata.obs[groupby].cat.categories)

    # Warn if there are more than 50 categories
    if n_categories > 50:
        warnings.warn(
            f"The '{groupby}' column has {n_categories} groups, which may result in "
            "slow runtimes. Ensure that {groupby} is not continuous data.", 
            UserWarning
        )

    # Check if rank_genes_groups has already been run
    if 'rank_genes_groups' not in adata.uns or adata.uns['rank_genes_groups']['params']['groupby'] != groupby:
        print(
            f"rerunning diffexp analysis because not found in adata.uns for "
            f"adata.obs['{groupby}']. (run before annotating to avoid this)"
        )
        sc.tl.rank_genes_groups(adata, groupby, method='t-test')

    rank_genes_groups = adata.uns['rank_genes_groups']
    clusters = rank_genes_groups['names'].dtype.names

    return rank_genes_groups, clusters

def _validate_metadata_columns(
    adata: AnnData,
    tissue_col: str | None = None,
    cell_type_col: str | None = None
) -> tuple | None:
    """
    Validates the existence of metadata columns and returns which ones are valid.
    
    Parameters
    ----------
    adata
        An :class:`AnnData`.

    tissue_col
        Name of tissue column

    cell_type_col
        Name of cell type column
        
    Returns
    -------
    A :class:`tuple` of the form ``(tissue_col, cell_type_col)``
    or ``None`` for invalid columns
    """
    if tissue_col and tissue_col not in adata.obs.columns:
        warnings.warn(
            f"Tissue of origin column '{tissue_col}' not found in adata.obs, "
            "will not consider tissue of origin for cell type annotation.",
            UserWarning
        )
        tissue_col = None

    if cell_type_col and cell_type_col not in adata.obs.columns:
        warnings.warn(
            f"Cell type of origin column '{cell_type_col}' not found in adata.obs, "
            "will not consider cell type of origin for annotation.",
            UserWarning
        )
        cell_type_col = None

    return tissue_col, cell_type_col

def _get_cluster_metadata(
    adata: AnnData,
    clusters: list[str | int] | tuple[str | int],
    groupby: str,
    tissue_col: str | None = None,
    cell_type_col: str | None = None
) -> tuple:
    """
    Gets tissue and cell type metadata for each cluster if available.

    Parameters
    ----------
    adata
        An :class:`AnnData`.

    clusters
        Sequence of cluster names

    groupby
        The column name used for grouping

    tissue_col
        Name of tissue column

    cell_type_col
        Name of cell type column

    Returns
    -------
    A :class:`tuple` of the form ``(cluster_to_tissue, cluster_to_cell_type)``
    """
    cluster_to_tissue = {}
    cluster_to_cell_type = {}

    if tissue_col or cell_type_col:
        for cluster in clusters:
            mask = adata.obs[groupby] == cluster

            if tissue_col:
                tissues = adata.obs.loc[mask, tissue_col].unique()
                cluster_to_tissue[cluster] = tissues[0] if len(tissues) == 1 else ", ".join(tissues)

            if cell_type_col:
                cluster_to_cell_type[cluster] = adata.obs.loc[mask, cell_type_col].unique().tolist()

    return cluster_to_tissue, cluster_to_cell_type

def _update_adata_and_create_results(
    adata: AnnData,
    cell_type_annotations: dict,
    groupby: str,
    new_label_column: str,
    n_top_genes: int,
    top_genes_lists: dict | list | None = None
) -> DataFrame:
    """
    Updates the AnnData object with annotations and creates the results DataFrame.

    Parameters
    ----------
    adata
        An :class:`AnnData`.

    cell_type_annotations
        Mapping of clusters to their annotations

    groupby
        The column name used for grouping

    new_label_column
        Name of the new annotation column

    n_top_genes
        Number of top genes used

    top_genes_lists
        Top genes for each cluster
    Returns
    -------
    Results DataFrame
    """
    # Create results list
    results = []
    for cluster, annotation in cell_type_annotations.items():
        result = {
            groupby: cluster,
            new_label_column: annotation,
        }

        if isinstance(top_genes_lists, dict):
            result[f"top_{n_top_genes}_genes"] = top_genes_lists[cluster]
        elif isinstance(top_genes_lists, list):
            cluster_idx = list(cell_type_annotations.keys()).index(cluster)
            result[f"top_{n_top_genes}_genes"] = top_genes_lists[cluster_idx]

        results.append(result)

    # Update adata
    adata.obs[new_label_column] = adata.obs[groupby].map(cell_type_annotations)
    convert_obs_col_to_category(adata, new_label_column)

    return pd.DataFrame(results)

def ai_annotate(
    func: callable,
    adata: AnnData,
    groupby: str,
    n_top_genes: int,
    new_label_column: str,
    tissue_of_origin_col: str = None,
    **kwargs,
) -> DataFrame:
    """
    Annotate clusters based on the top marker genes for each cluster.

    This uses marker genes for each cluster and applies func to 
    determine the label for each cluster based on the top n
    marker genes. The results are added to the AnnData 
    object and returned as a DataFrame.

    If rank_genes_groups hasn't been run on the adata, 
    this function will automatically run ``sc.tl.rank_genes_groups``

    Parameters
    ------------
    func
        A function that takes ``gene_list`` 
        **:** :class:`list` **[** :class:`str` **]** and returns ``annotation`` **:** :class:`str`.

    adata
        An :class:`AnnData` object.

    groupby
        Column in ``adata.obs`` to group by for differential expression analysis.

    n_top_genes
        The number of top marker genes to consider for each cluster.

    new_label_column
        The name of the new column in ``adata.obs`` where the annotations will be stored.

    tissue_of_origin_col
        Name of a column in ``adata.obs`` that contains 
        the tissue of orgin. Used to provide context to the LLM.

    **kwargs
        additional kwargs passed to ``func``

    Returns
    ---------
    A ``pd.DataFrame`` with a column for the top marker genes for each cluster.

    Notes
    -------
    This function also modifies the input ``adata`` in place, 
    adding annotations to ``adata.obs[new_label_col]``
    """
    # Validate and prepare rank genes data
    rank_genes_groups, clusters = _validate_and_prepare_rank_genes(adata, groupby)

    # Validate metadata columns
    tissue_of_origin_col, _ = _validate_metadata_columns(adata, tissue_col=tissue_of_origin_col)

    # Get cluster metadata
    cluster_to_tissue, _ = _get_cluster_metadata(
        adata, clusters, groupby, tissue_col=tissue_of_origin_col
    )

    # Process each cluster
    cell_type_annotations = {}
    top_genes_dict = {}

    for cluster in clusters:
        if tissue_of_origin_col:
            kwargs['tissue'] = cluster_to_tissue[cluster]

        top_genes = list(rank_genes_groups['names'][cluster][:n_top_genes])
        top_genes_dict[cluster] = top_genes

        annotation = func(top_genes, **kwargs)
        cell_type_annotations[cluster] = annotation

    return _update_adata_and_create_results(
        adata, cell_type_annotations, groupby, new_label_column, 
        n_top_genes, top_genes_dict
    )

def ai_annotate_by_comparison(
    func :callable,
    adata: AnnData,
    groupby: str,
    n_top_genes: int,
    new_label_column: str,
    cell_type_of_origin_col: str = None,
    tissue_of_origin_col: str = None,
    **kwargs,
) -> DataFrame:
    """
    Annotate clusters based on the top marker genes for each cluster, 
    in the context of the other clusters' marker genes.

    This uses marker genes for each cluster and applies func to 
    determine the label for each cluster based on the top n
    marker genes. The results are added to the AnnData object 
    and returned as a DataFrame.

    If rank_genes_groups hasn't been run on the adata, this 
    function will automatically run ``sc.tl.rank_genes_groups``

    Parameters
    -------------
    func
        A function that takes ``gene_lists`` **:** :class:`list` 
        **[** :class:`list` **[** :class:`str` **] ]** and

        returns ``annotations`` **:** :class:`list` **[** :class:`str` **]**, 
        one for each :class:`list` of genes in ``gene_lists``.

    adata
        An :class:`AnnData` object.

    groupby
        Column in ``adata.obs`` to group by for differential expression analysis.

    n_top_genes
        The number of top marker genes to consider for each cluster.

    new_label_column
        The name of the new column in ``adata.obs`` where the annotations will be stored.

    cell_type_of_origin_col
        Name of a column in ``adata.obs`` that contains the 
        cell type of orgin. Used for context to the LLM.

    tissue_of_origin_col
        Name of a column in ``adata.obs`` that contains the 
        tissue of orgin. Used to provide context to the LLM.

    **kwargs
        additional kwargs passed to ``func``

    Returns
    --------
    A ``pd.DataFrame`` with a column for the top marker genes for each cluster.

    Notes
    -------
    This function also modifies the input ``adata`` in place, 
    adding annotations to ``adata.obs[new_label_col]``
    """
    # Validate and prepare rank genes data
    rank_genes_groups, clusters = _validate_and_prepare_rank_genes(adata, groupby)

    # Validate metadata columns
    tissue_of_origin_col, cell_type_of_origin_col = _validate_metadata_columns(
        adata, tissue_of_origin_col, cell_type_of_origin_col
    )

    # Get cluster metadata
    cluster_to_tissue, cluster_to_cell_type = _get_cluster_metadata(
        adata, clusters, groupby, tissue_of_origin_col, cell_type_of_origin_col
    )

    # Get top genes for all clusters
    top_genes = [list(rank_genes_groups['names'][cluster][:n_top_genes]) for cluster in clusters]

    # Prepare kwargs with metadata
    if tissue_of_origin_col:
        kwargs['tissues'] = [cluster_to_tissue[cluster] for cluster in clusters]
    if cell_type_of_origin_col:
        kwargs['cell_types'] = [cluster_to_cell_type[cluster] for cluster in clusters]

    # Get annotations for all clusters at once
    annotations = func(top_genes, **kwargs)

    # Create cell type annotations dictionary
    cell_type_annotations = dict(zip(clusters, annotations))

    return _update_adata_and_create_results(
        adata, cell_type_annotations, groupby, new_label_column, 
        n_top_genes, top_genes
    )



"""
Functions to read spatial transcriptomic data.
"""

from typing import Literal

import numpy as np
import scanpy as sc
import anndata as ad
import pandas as pd

from scipy.sparse import dok_matrix
from scipy.sparse import csr_matrix
from anndata import AnnData


def read_transcript_coords(
    file_path: str,
    platform: Literal["Merscope", "Xenium"] = None
) -> pd.DataFrame:
    """
    Reads the transcript locations from a CSV file and ensures it contains the necessary columns.

    Parameters
    ----------
    file_path
        The path to the CSV file.

    platform
        The platform type.

    Returns
    -------
    A pandas DataFrame containing the transcript locations.

    Raises
    ------
    ValueError
        If the required columns are not present in the file or if the file format is unsupported.
    """
    print("reading data")

    if not file_path.endswith(".csv"):
        raise ValueError("Unsupported file format. Please provide a CSV file.")

    df = pd.read_csv(file_path)

    required_columns_merscope = {"global_x", "global_y", "gene"}
    required_columns_xenium = {"feature_name", "x_location", "y_location"}

    if platform == "Merscope":
        if not required_columns_merscope.issubset(df.columns):
            raise ValueError(
                f"The file must contain the following columns for Merscope: {required_columns_merscope}"
            )
    elif platform == "Xenium":
        if not required_columns_xenium.issubset(df.columns):
            raise ValueError(
                f"The file must contain the following columns for Xenium: {required_columns_xenium}"
            )
        df = df.rename(
            columns={
                "feature_name": "gene",
                "x_location": "global_x",
                "y_location": "global_y",
            }
        )
    else:
        raise ValueError(
            "Unsupported platform. Please provide either 'Merscope' or 'Xenium' as the platform."
        )

    return df


def get_steps_and_coords(
    df: pd.DataFrame,
    box_size: int,
    step_size: int
) -> tuple[int, int, list[list[int]]]:
    """
    Computes the number of steps and the top-left coordinates of each box.

    Parameters
    ----------
    df
        The data containing 'global_x' and 'global_y' columns.

    box_size
        The size of the box.

    step_size
        The step size.

    Returns
    -------
    A tuple containing the number of steps in x and y directions, and the list of top-left coordinates of each box.

    Raises
    ------
    ValueError
        If the box size is larger than the image dimensions.
    """
    print("getting steps and coords")
    min_x, max_x = df["global_x"].min(), df["global_x"].max()
    min_y, max_y = df["global_y"].min(), df["global_y"].max()

    x_steps = int((max_x - min_x - box_size) / step_size) + 2
    y_steps = int((max_y - min_y - box_size) / step_size) + 2

    if x_steps < 0:
        raise ValueError("box size is larger than image")

    coords_top_left = [
        [min_x + step_size * i, min_y + step_size * j]
        for i in range(0, x_steps)
        for j in range(0, y_steps)
    ]

    return x_steps, y_steps, coords_top_left


def populate_sparse_array(
    df: pd.DataFrame,
    genes: np.array,
    step_size: int
) -> csr_matrix:
    """
    Populates a sparse array with gene counts.

    Parameters
    ----------
    df
        The data containing 'global_x', 'global_y', and 'gene' columns.

    genes
        The unique genes.

    step_size
        The step size.

    Returns
    -------
    The sparse matrix with gene counts.
    """
    num_boxes_x = int((df["global_x"].max() - df["global_x"].min()) // step_size) + 1
    num_boxes_y = int((df["global_y"].max() - df["global_y"].min()) // step_size) + 1
    num_boxes = num_boxes_x * num_boxes_y

    sparse_array = dok_matrix((num_boxes, len(genes)), dtype=np.int32)
    gene_to_index = {gene: idx for idx, gene in enumerate(genes)}

    min_x, min_y = df["global_x"].min(), df["global_y"].min()
    df["box_x"] = ((df["global_x"] - min_x) // step_size).astype(int)
    df["box_y"] = ((df["global_y"] - min_y) // step_size).astype(int)

    for (box_x, box_y), box_df in df.groupby(["box_x", "box_y"]):
        index = box_x * num_boxes_y + box_y
        if 0 <= index < num_boxes:
            gene_counts = box_df["gene"].value_counts()
            for gene, count in gene_counts.items():
                if gene in gene_to_index:
                    sparse_array[index, gene_to_index[gene]] = count

    return sparse_array.tocsr()


def process_gene_counts(
    file_path: str,
    box_size: int,
    step_size: int,
    platform: Literal["Merscope", "Xenium"] = None,
) -> tuple[csr_matrix, np.array, list[list[int]]]:
    """
    Processes the gene counts from the CSV file.

    Parameters
    ----------
    file_path
        The path to the CSV file.

    box_size
        The size of the box.

    step_size
        The step size.

    platform
        The platform type.

    Returns
    -------
    A tuple containing the sparse matrix, unique genes, and list of top-left coordinates of each box.
    """
    df = read_transcript_coords(file_path, platform=platform)
    print("processing gene counts")
    genes = df["gene"].unique()
    _, _, coords_top_left = get_steps_and_coords(df, box_size, step_size)
    sparse_array = populate_sparse_array(df, genes, step_size)
    return sparse_array, genes, coords_top_left


def create_anndata(
    sparse_array: csr_matrix,
    genes: np.array,
    coords_top_left: list[list[int]]
) -> AnnData:
    """
    Creates an :clsas:`AnnData` object from the sparse matrix and coordinates.

    Parameters
    ----------
    sparse_array
        The sparse matrix with gene counts.

    genes
        The unique genes.

    coords_top_left
        The list of top-left coordinates of each box.

    Returns
    -------
    An :class:`AnnData` object containing the gene counts and metadata.
    """
    print("creating anndata")

    # Create the AnnData object, setting the sparse matrix as the X attribute
    adata = ad.AnnData(X=sparse_array.tocsr(), var={"gene_symbols": genes})

    # Set the gene symbols as the index of .var
    adata.var.index = adata.var["gene_symbols"]

    # Add the top-left coordinates of each box ("cell") as .obs
    adata.obs = pd.DataFrame(
        coords_top_left,
        columns=["global_x_topleft", "global_y_topleft"],
        index = [f"region_{i}" for i in range(len(coords_top_left))]
    )

    return adata


def add_blank_image_to_adata(
    adata: AnnData,
    platform: Literal["Merscope", "Xenium"] = "Merscope"
) -> AnnData:
    """
    Adds a dummy image to the AnnData object based on the platform specifications.

    Parameters
    ----------
    adata
        The :class:`AnnData` object to which the image will be added.

    platform
        The platform from which the data originates.

    Returns
    -------
    ``adata`` updated with a blank image and spatial data.
    """
    adata.obsm["spatial"] = adata.obs[
        ["global_x_topleft", "global_y_topleft"]
    ].to_numpy()

    max_x = int(adata.obs["global_x_topleft"].max())
    max_y = int(adata.obs["global_y_topleft"].max())

    dummy_image = np.ones((max_y + 1, max_x + 1, 3))

    if platform in ["Merscope", "Xenium"]:
        adata.uns["spatial"] = {
            "library_id": {
                "images": {"hires": dummy_image, "lowres": dummy_image},
                "scalefactors": {
                    "tissue_hires_scalef": 1.0,
                    "tissue_lowres_scalef": 1.0,
                    "spot_diameter_fullres": 16,
                },
            }
        }

    return adata


def build_adata_from_transcript_positions(
    paths_dict: dict[str, str],
    box_size: int = 16,
    step_size: int = 16,
    platform: Literal["Merscope", "Xenium"] = "Merscope",
) -> None:
    """
    Builds an :class:`AnnData` object from a ``detected_transcripts.csv`` (Merscope) or 
    ``transcripts.csv`` (Xenium) file and saves it to a specified output path. 
    These are the files output by most spatial transcriptomic platforms, 
    including Visium, Visium HD, Xenium, and Merscope.

    Parameters
    ----------
    paths_dict
        A dictionary with input paths as keys and output paths as values.

    box_size
        The size of the box.

    step_size
        The step size.

    platform
        The platform used, either "Merscope" (default) or "Xenium".

    Returns
    -------
    None

    Notes
    -----
    Writes the processed :class:`AnnData` objects to the specified output paths.

    Example
    -------

    ..code-block:: python

        paths_dict = {
            'input_path1.csv': 'output_path1.h5ad',
            'input_path2.csv': 'output_path2.h5ad'
        }

        build_adata_from_transcript_positions(paths_dict)
    """
    for input_path, output_path in paths_dict.items():
        sparse_array, genes, coords_top_left = process_gene_counts(
            input_path, box_size, step_size, platform=platform
        )
        adata = create_anndata(sparse_array, genes, coords_top_left)
        adata = add_blank_image_to_adata(adata, platform=platform)
        adata.write(output_path)


def build_adata_from_visium(
    paths_dict: dict[str, str],
    hd: bool = False
) -> None:
    """
    Processes Visium data from input directories and saves the processed AnnData objects to specified output paths.

    Parameters
    ----------
    paths_dict
        A dictionary with input directories as keys and output file paths as values.

    hd
        If ``True``, converts ``'spatial'`` to float and ``'obs'`` columns to integers.

    Returns
    -------
    None

    Notes
    -----
    Writes the processed :class:`AnnData` objects to the specified output paths.

    Example
    -------

    .. code-block:: python

        paths_dict = {
            'input_dir1': 'output_path1.h5ad',
            'input_dir2': 'output_path2.h5ad'
        }

        build_adata_from_visium(paths_dict, hd=True)
    """
    for input_dir, output_path in paths_dict.items():
        adata = sc.read_visium(input_dir)

        if hd:
            # Run fixes to sc.read_visium for visium HD data
            adata.obsm["spatial"] = adata.obsm["spatial"].astype(float)
            adata.obs = adata.obs.astype(int)

        adata.write(output_path)



"""
This module contains functions that process labels within a single column.
"""

from .in_adata_obs import (
    simplify_obs_column,
    create_label_hierarchy,
    map_cell_type_labels_to_simplified_set,

)

from .in_adata_var import (
    simplify_var_index,
    map_gene_labels_to_simplified_set,

)

__all__ = [
    #in_adata_obs.py
    "simplify_obs_column",
    "create_label_hierarchy",
    "map_cell_type_labels_to_simplified_set",

    #in_adata_var.py
    "simplify_var_index",
    "map_gene_labels_to_simplified_set",

]



"""
This module contains visualizations functions for gene annotation.
"""
import textwrap
import matplotlib.pyplot as plt

from pandas import DataFrame

def annotate_gene_groups_with_ai_biological_process(
    plot_obj: dict,
    df: DataFrame,
    groupby: str,
    adt_key: tuple[str,...] | None = None
) -> None:
    """
    Annotate gene groups on a heatmap with their biological process labels.

    This function adds annotations below the x-axis of a heatmap generated by
    :func:`sc.pl.rank_genes_groups_heatmap`, labeling gene groups with their associated
    biological process descriptions from :func:`adt.ai_annotate_biological_process`.

    Parameters
    ----------
    plot_obj
        Output from `sc.pl.rank_genes_groups_heatmap`, containing matplotlib axes,
        specifically with the key 'heatmap_ax' representing the heatmap axis.

    df
        Output from :func:`adt.ai_annotate_biological_process`, containing gene group information
        with the following columns:
            - ``'groupby'``: Identifier for each gene group.
            - ``'top_10_genes'``: List of gene names in the group.
            - ``'ai_biological_process'``: Biological process label for the group.

    group_by
        The grouping column name.

    adt_key
        Used by :func:`adata_dict_fapply` and :func:`adata_dict_fapply_return` when passing this function.

    Returns
    -------
    ``None``

    See Also
    ----------
    :func:`sc.pl.rank_genes_groups_heatmap` : To generate the ``plot_obj``.
    :func:`adt.ai_annotate_biological_process` : To annotate the gene sets.
    """

    if adt_key:
        print(adt_key)

    heatmap_ax = plot_obj['heatmap_ax']

    # Get the genes and their positions from the x-axis
    xticks = heatmap_ax.get_xticks()
    xlabels = [tick.get_text() for tick in heatmap_ax.get_xticklabels()]
    gene_positions = {gene: xtick for gene, xtick in zip(xlabels, xticks)}

    group_positions = {}
    for idx, row in df.iterrows():
        group = row[groupby]
        genes_in_group = row['top_10_genes']
        # Filter genes that are actually in the plot
        positions = [gene_positions[gene] for gene in genes_in_group if gene in gene_positions]
        if positions:
            start_pos = min(positions)
            end_pos = max(positions)
            center_pos = (start_pos + end_pos) / 2.0
            group_positions[group] = {'start': start_pos, 'end': end_pos, 'center': center_pos}

    # Adjust the figure to make space for annotations
    fig = heatmap_ax.figure
    bottom_margin = 0.18  # Adjust this value if more space is needed
    fig.subplots_adjust(bottom=bottom_margin)

    # Annotate each group with the ai_biological_process label
    for group, pos_dict in group_positions.items():
        process_label = df.loc[df[groupby] == group, 'ai_biological_process'].values[0]
        # Wrap the text to fit within the group width
        # Convert group width from data to display coordinates
        inv = heatmap_ax.transData.inverted()
        start_display = heatmap_ax.transData.transform((pos_dict['start'], 0))
        end_display = heatmap_ax.transData.transform((pos_dict['end'], 0))
        group_display_width = end_display[0] - start_display[0]
        # Convert display width to figure coordinates
        start_fig = fig.transFigure.inverted().transform(start_display)
        end_fig = fig.transFigure.inverted().transform(end_display)
        group_fig_width = end_fig[0] - start_fig[0]
        # Estimate character width in figure coordinates
        char_width = 0.006  # Adjust this value based on font size and figure size
        max_chars = int((group_fig_width - 0.15) / char_width)  # Subtract padding
        max_chars = max(max_chars, 0.5)
        wrapped_label = textwrap.fill(process_label, width=max_chars)

        # Get the center position in figure coordinates
        center_display = heatmap_ax.transData.transform((pos_dict['center'], 0))
        center_fig = fig.transFigure.inverted().transform(center_display)

        # Place the text below the x-axis labels
        y_pos = 0.02  # Position in figure coordinates, adjust as needed
        fig.text(center_fig[0], y_pos, wrapped_label, ha='center', va='top', fontsize=10, fontweight='bold')

    # Redraw the plot to show the annotations
    fig.canvas.draw()

    plt.figure(plot_obj['gene_groups_ax'].figure)  # Select the figure associated with the heatmap axis
    plt.show()



"""
This module contains the functions necessary to read :class:`AdataDict` objects from adata on disk.
"""

import os
import json
from collections import Counter

import anndata as ad
import scanpy as sc

from .adata_dict import AdataDict, to_nested_tuple

def read_adata_dict(
    directory: str,
) -> AdataDict:
    """
    Read an :class:`AdataDict` from a previously saved :class:`AdataDict`. 
    To write an :class:`AdataDict` see :func:`~write_adata_dict`.

    Parameters
    -----------
    directory
        Base directory where the ``.h5ad`` files and hierarchy file are located.

    Returns
    --------
    An :class:`AdataDict` reconstructed from the saved files.

    Raises
    ------
    FileNotFoundError
        If the directory or required metadata files don't exist.
    ValueError
        If the metadata files are corrupted or in an unexpected format.

    Examples
    ---------
    **Case 1: Flat hierarchy**

    .. code-block:: bash

        directory/
        ├── adata_dict.hierarchy.json
        ├── adata_dict.db.json
        ├── <file_prefix>Donor1_Tissue1.h5ad
        ├── <file_prefix>Donor1_Tissue2.h5ad
        └── <file_prefix>Donor2_Tissue1.h5ad

    The reconstructed AdataDict will be:

    .. code-block:: python

        print(adata_dict)
        > {
        >     ("Donor1", "Tissue1"): adata_d1_t1,
        >     ("Donor1", "Tissue2"): adata_d1_t2,
        >     ("Donor2", "Tissue1"): adata_d2_t1,
        > }

    **Case 2: Nested hierarchy**

    .. code-block:: bash

        directory/
        ├── adata_dict.hierarchy.json
        ├── adata_dict.db.json
        ├── Donor1/
        │   ├── <file_prefix>Tissue1.h5ad
        │   └── <file_prefix>Tissue2.h5ad
        └── Donor2/
            └── <file_prefix>Tissue1.h5ad

    The reconstructed AdataDict will be:

    .. code-block:: python

        print(adata_dict)
        > {
        >     ("Donor1",): {
        >         ("Tissue1",): adata_d1_t1,
        >         ("Tissue2",): adata_d1_t2,
        >     },
        >     ("Donor2",): {
        >         ("Tissue1",): adata_d2_t1,
        >     },
        > }

    **Case 3: Nested hierarchy with multiple indices at the deepest level**

    .. code-block:: bash

        directory/
        ├── adata_dict.hierarchy.json
        ├── adata_dict.db.json
        ├── Donor1/
        │   ├── <file_prefix>Tissue1_CellType1.h5ad
        │   ├── <file_prefix>Tissue1_CellType2.h5ad
        │   └── <file_prefix>Tissue2_CellType3.h5ad
        └── Donor2/
            └── <file_prefix>Tissue1_CellType1.h5ad

    The reconstructed AdataDict will be:

    .. code-block:: python

        print(adata_dict)
        > {
        >     ("Donor1",): {
        >         ("Tissue1", "CellType1"): adata_d1_t1_c1,
        >         ("Tissue1", "CellType2"): adata_d1_t1_c2,
        >         ("Tissue2", "CellType3"): adata_d1_t2_c3,
        >     },
        >     ("Donor2",): {
        >         ("Tissue1", "CellType1"): adata_d2_t1_c1,
        >     },
        > }

    See Also
    ---------
    :func:`~write_adata_dict` : To write an :class:`AdataDict`
    """
    # Check if directory exists
    if not os.path.exists(directory):
        raise FileNotFoundError(f"Directory {directory} does not exist")

    # Read the hierarchy
    hierarchy_path = os.path.join(directory, "adata_dict.hierarchy.json")
    if not os.path.exists(hierarchy_path):
        raise FileNotFoundError(f"Hierarchy file not found in {directory}")

    with open(hierarchy_path, "r", encoding="utf-8") as f:
        hierarchy = json.load(f)

    # Read the database mapping files to keys
    db_path = os.path.join(directory, "adata_dict.db.json")
    if not os.path.exists(db_path):
        raise FileNotFoundError(f"Database file not found in {directory}")

    with open(db_path, "r", encoding="utf-8") as f:
        db_entries = json.load(f)

    # Helper function to create nested dictionary from key path
    def create_nested_dict(data_dict: dict, key_path: tuple, value: any) -> None:
        """Create nested dictionary structure from a key path and value."""
        # For flat hierarchies, use the full key_path as a single tuple key
        if not any(isinstance(x, list) for x in hierarchy):
            data_dict[key_path] = value
            return

        # For nested hierarchies, we need to handle each level
        current = data_dict
        # Split the key_path according to the hierarchy structure
        current_pos = 0
        for level in hierarchy:
            if isinstance(level, list):
                # This level contains the remaining elements
                remaining_len = len(level)
                level_key = key_path[current_pos:current_pos + remaining_len]
                current[level_key] = value
                break
            else:
                # This level is a single element
                level_key = (key_path[current_pos],)
                if level_key not in current:
                    current[level_key] = {}
                current = current[level_key]
                current_pos += 1

    # Initialize the data dictionary
    data_dict: dict = {}

    # Read each AnnData file and reconstruct the dictionary
    for entry in db_entries:
        file_path = os.path.join(directory, entry["file_path"])
        if not os.path.exists(file_path):
            raise FileNotFoundError(f"AnnData file not found: {file_path}")

        key = tuple(entry["key"])  # Convert key from list to tuple
        adata = sc.read_h5ad(file_path)
        create_nested_dict(data_dict, key, adata)

    # Create and return the AdataDict
    return AdataDict(data_dict, hierarchy=to_nested_tuple(hierarchy))


def read_adata_dict_from_h5ad(
    paths: str | list[str],
    *,
    keys: list[str] | None = None,
) -> AdataDict:
    """
    Read ``.h5ad`` files from a list of paths and return them in a dictionary.

    For each element in the provided list of paths, if the element is a directory, 
    it reads all ``.h5ad`` files in that directory. If the element is an ``.h5ad`` file, 
    it reads the file directly.

    For auto-generated keys, if there are duplicate filenames, the function will 
    include parent directory names from right to left until keys are unique. 
    For example, ``dat/heart/fibroblast.h5ad`` would generate the key ``('heart', 'fibroblast')``
    if disambiguation is needed.

    Parameters
    ------------
    paths
        A string path or list of paths to directories or ``.h5ad`` files.

    keys
        A list of strings to use as keys for the adata_dict. If provided, must be equal 
        in length to the number of ``.h5ad`` files read.

    Returns
    -------
    An :class:`AdataDict`.
    """

    if isinstance(paths, str):
        paths = [paths]

    adata_dict = {}
    file_paths = []

    # First, collect all file paths
    for path in paths:
        if os.path.isdir(path):
            for file in os.listdir(path):
                if file.endswith(".h5ad"):
                    file_paths.append(os.path.join(path, file))
        elif path.endswith(".h5ad"):
            file_paths.append(path)

    # Check if provided keys match the number of files
    if keys is not None:
        if len(keys) != len(file_paths):
            raise ValueError(
                f"Number of provided keys ({len(keys)}) does not match the number of .h5ad files ({len(file_paths)})"
            )
        # Check for uniqueness in provided keys
        key_counts = Counter(keys)
        duplicates = [k for k, v in key_counts.items() if v > 1]
        if duplicates:
            raise ValueError(f"Duplicate keys found: {duplicates}")
        # Convert provided keys to tuples
        tuple_keys = [tuple(k) if isinstance(k, (list, tuple)) else (k,) for k in keys]
    else:
        # Generate keys from paths
        base_names = [os.path.splitext(os.path.basename(fp))[0] for fp in file_paths]

        # Start with just the base names
        tuple_keys = [(name,) for name in base_names]

        # Keep extending paths to the left until all keys are unique
        while len(set(tuple_keys)) != len(tuple_keys):
            new_tuple_keys = []
            for i, file_path in enumerate(file_paths):
                path_parts = os.path.normpath(file_path).split(os.sep)
                # Find the current key's elements in the path
                current_key = tuple_keys[i]
                current_idx = (
                    len(path_parts) - 1 - len(current_key)
                )  # -1 for zero-based index
                # Add one more path element to the left if possible
                if current_idx > 0:
                    new_key = (path_parts[current_idx - 1],) + current_key
                else:
                    new_key = current_key
                new_tuple_keys.append(new_key)
            tuple_keys = new_tuple_keys

            # Safety check - if we've used all path components and still have duplicates
            if all(
                len(key) == len(os.path.normpath(fp).split(os.sep))
                for key, fp in zip(tuple_keys, file_paths)
            ):
                raise ValueError("Unable to create unique keys even using full paths")

    # Process the files with the finalized tuple keys
    for i, file_path in enumerate(file_paths):
        adata_dict[tuple_keys[i]] = ad.read_h5ad(file_path)

    return adata_dict


def read(
    directory_list: str | list[str],
    *,
    keys: list[str] | None = None,
) -> AdataDict:
    """
    Process a list of directories or file paths to read AnnData objects. 

    For each directory, if a `.hierarchy` file is found, the directory is processed 
    with `read_adata_dict`. Otherwise, the highest-level directory is processed with 
    `read_adata_dict_from_h5ad`. Subdirectories of a directory containing a `.hierarchy` 
    file are not processed with `read_adata_dict_from_h5ad`.

    Parameters
    ------------
    directory_list
        String or list of strings, paths to directories or `.h5ad` files.

    keys
        List of strings that will be used as keys for the resulting dictionary.

    Returns
    -------
    An :class:`AdataDict` of all AnnData objects.
    """
    if isinstance(directory_list, str):
        directory_list = [directory_list]

    adata_dict = {}

    # Set to keep track of directories that have been processed with read_adata_dict
    hierarchy_dirs = set()

    # List to collect .h5ad files to process
    h5ad_files = []

    # Function to find all directories containing adata_dict.hierarchy files
    def find_hierarchy_dirs(dir_path):
        for root, dirs, files in os.walk(dir_path):
            if "adata_dict.hierarchy" in files:
                hierarchy_dirs.add(root)
                # Do not traverse subdirectories of directories with hierarchy files
                dirs[:] = []
            else:
                # Continue traversing subdirectories
                pass

    # First, process the input paths to find hierarchy directories and collect .h5ad files
    for path in directory_list:
        if os.path.isfile(path):
            if path.endswith(".h5ad"):
                h5ad_files.append(path)
        elif os.path.isdir(path):
            # Find hierarchy directories
            find_hierarchy_dirs(path)
        else:
            raise ValueError(f"Path {path} is neither a file nor a directory.")

    # Process directories with hierarchy files using read_adata_dict
    for h_dir in hierarchy_dirs:
        adata_dict.update(read_adata_dict(h_dir))

    # Build a set of directories to exclude (hierarchy_dirs and their subdirectories)
    exclude_dirs = set()
    for h_dir in hierarchy_dirs:
        for root, dirs, files in os.walk(h_dir):
            exclude_dirs.add(root)

    # Function to collect .h5ad files not under exclude_dirs
    def collect_h5ad_files(dir_path):
        for root, dirs, files in os.walk(dir_path):
            # Skip directories under exclude_dirs
            if any(
                os.path.commonpath([root, excl_dir]) == excl_dir
                for excl_dir in exclude_dirs
            ):
                dirs[:] = []
                continue
            for file in files:
                if file.endswith(".h5ad"):
                    h5ad_files.append(os.path.join(root, file))

    # Collect .h5ad files from directories not containing hierarchy files
    for path in directory_list:
        if os.path.isdir(path):
            collect_h5ad_files(path)

    # Process the collected .h5ad files using read_adata_dict_from_h5ad
    if h5ad_files:
        adata_dict.update(read_adata_dict_from_h5ad(h5ad_files, keys=keys))

    return adata_dict


"""
This module contains the functions necessary to build AdataDict objects from adata in memory.
"""

import itertools

import pandas as pd
from anndata import AnnData

from .adata_dict import AdataDict


def build_adata_dict(
    adata: AnnData,
    strata_keys: list[str],
    *,
    desired_strata: list | dict | None = None,
) -> AdataDict:
    """
    Build a dictionary of :class:`AnnData` objects split 
    by the ``obs`` columns specified in ``strata_keys``.

    Parameters
    ------------
    adata
        The input :class:`AnnData`.

    strata_keys
        List of column names in ``adata.obs`` to use for stratification.

    desired_strata
        List of desired strata tuples or a dictionary where keys 
        are the col names supplied in ``strata_keys``, and values
        are lists of desired strata. If None (Default), all combinations of categories
        in ``adata.obs[strata_keys]`` will be used.

    Returns
    --------
    Flat dictionary of class :class:`AdataDict` where keys 
    are strata tuples and values are corresponding :class:`AnnData` subsets.

    Raises
    ------
    ValueError
        If `desired_strata` is neither a list nor a dictionary of lists.

    Examples
    --------
        Case 1: Take all groups
            >>> import pandas as pd
            >>> from anndata import AnnData
            >>> # Create an example AnnData object
            >>> adata = AnnData(obs=pd.DataFrame({
            ...     "Donor": ["Donor1", "Donor1", "Donor2"],
            ...     "Tissue": ["Tissue1", "Tissue2", "Tissue1"]
            ... }))
            >>> strata_keys = ["Donor", "Tissue"]
            >>> adata_dict = build_adata_dict(adata, strata_keys)
            >>> print(adata_dict)
            {
                ("Donor1", "Tissue1"): adata_d1_t1,
                ("Donor1", "Tissue2"): adata_d1_t2,
                ("Donor2", "Tissue1"): adata_d2_t1,
            }
        Case 2: Take only some groups
            >>> import pandas as pd
            >>> from anndata import AnnData
            >>> # Create an example AnnData object
            >>> adata = AnnData(obs=pd.DataFrame({
            ...     "Donor": ["Donor1", "Donor1", "Donor2"],
            ...     "Tissue": ["Tissue1", "Tissue2", "Tissue1"]
            ... }))
            >>> strata_keys = ["Donor", "Tissue"]
            >>> desired_strata = {
                                    "Donor": ["Donor1"], #Take only Donor1
                                    "Tissue": ["Tissue1", "Tissue2"] #Take Tissue1 and Tissue2
            }
            >>> adata_dict = build_adata_dict(adata, strata_keys, desired_strata=desired_strata)
            >>> print(adata_dict)
            {
                ("Donor1", "Tissue1"): adata_d1_t1,
                ("Donor1", "Tissue2"): adata_d1_t2
            }
    """
    if desired_strata is None:
        # Generate all combinations of categories in adata.obs[strata_keys]
        all_categories = [adata.obs[key].cat.categories.tolist() for key in strata_keys]
        all_combinations = list(itertools.product(*all_categories))
        desired_strata = all_combinations
        return build_adata_dict_main(
            adata, strata_keys, desired_strata, print_missing_strata=False
        )

    if isinstance(desired_strata, list):
        # Ensure that desired_strata is a list of tuples
        if all(isinstance(item, str) for item in desired_strata):
            raise ValueError("desired_strata should be a list of tuples, not strings.")
        return build_adata_dict_main(adata, strata_keys, desired_strata)

    if isinstance(desired_strata, dict):
        # Generate all combinations of desired strata values across strata_keys
        all_combinations = itertools.product(
            *(desired_strata[key] for key in strata_keys)
        )
        desired_strata = list(all_combinations)
        return build_adata_dict_main(adata, strata_keys, desired_strata)

    raise ValueError(
        "desired_strata must be either a list of tuples or a dictionary of lists"
    )


def build_adata_dict_main(
    adata: AnnData,
    strata_keys: list[str],
    desired_strata: list[tuple],
    *,
    print_missing_strata: bool = True,
) -> AdataDict:
    """
    Function to build a dictionary of AnnData objects based on desired strata values.

    Parameters
    ------------
    adata
        Annotated data matrix.

    strata_keys
        List of column names in `adata.obs` to use for stratification.

    desired_strata
        List of desired strata tuples.

    print_missing_strata
        If True, print missing strata. Default is True.

    Returns
    -------
    Flat dictionary of class :class:`AdataDict` where 
    keys are strata tuples and values are corresponding AnnData subsets.
    """
    # Ensure that the strata columns are categorical
    for key in strata_keys:
        if not isinstance(adata.obs[key].dtype, pd.CategoricalDtype):
            adata.obs[key] = adata.obs[key].astype("category")

    # Group indices by combinations of strata_keys for efficient access
    groups = adata.obs.groupby(strata_keys, observed=False).indices

    # Adjust group keys to always be tuples
    if len(strata_keys) == 1:
        groups = {(k,): v for k, v in groups.items()}

    # Initialize the dictionary to store subsets
    adata_dict = {}

    # Iterate over desired strata (tuples) and extract subsets
    for stratum in desired_strata:
        if stratum in groups:
            indices = groups[stratum]
            adata_dict[stratum] = adata[indices].copy()
        else:
            if print_missing_strata:
                print(f"Warning: No {stratum} in data based on {strata_keys}.")

    # Create AdataDict and set hierarchy to strata_keys
    adata_dict = AdataDict(adata_dict, tuple(strata_keys))
    return adata_dict



"""
This module contains adata_dict wrappers for `scanpy`.
"""

import scanpy as sc

from anndata import AnnData

from anndict.adata_dict import (
    AdataDict,
    adata_dict_fapply,
    adata_dict_fapply_return,
    build_adata_dict,
    concatenate_adata_dict,
    check_and_create_stratifier,
)


def subsample_adata_dict(
    adata_dict: AdataDict,
    **kwargs
) -> None:
    """
    Subsamples each :class:`AnnData` in ``adata_dict`` using :func:`sc.pp.subsample`.
    
    Parameters
    -----------
    adata_dict
        An :class:`AdataDict`.

    kwargs
        Additional keyword arguments to pass to :func:`sc.pp.subsample`

    Returns
    --------
    None

    Notes
    -----
    The function modifies ``adata_dict`` in-place.
    """
    n_obs = kwargs.get('n_obs', None)
    fraction = kwargs.get('fraction', None)

    if n_obs is None and fraction is None:
        fraction = 1
        kwargs['fraction'] = fraction

    def subsample_adata(adata, **kwargs):
        if n_obs is None or adata.n_obs > n_obs:
            sc.pp.subsample(adata, **kwargs)

    adata_dict_fapply(adata_dict, subsample_adata, **kwargs)


def resample_adata(
    adata: AnnData,
    strata_keys: list[str],
    min_num_cells: int,
    n_largest_groups: int | None = None,
    **kwargs
) -> AnnData:
    """
    Resample ``adata`` based on specified strata keys and drop strata with fewer than the minimum number of cells.

    Parameters
    -----------
    adata
        An :class:`AnnData`.

    strata_keys
        List of column names in adata.obs to use for stratification.

    min_num_cells
        Minimum number of cells required to retain a stratum.

    kwargs
        Additional keyword arguments to pass to the subsample function.

    Returns
    --------
    Concatenated :class:`AnnData` object after resampling and filtering.

    Raises
    --------
    ValueError
        If any of the specified ``strata_keys`` do not exist in ``adata.obs``.
    """
    # Step 1: Create the strata key
    strata_key = check_and_create_stratifier(adata, strata_keys)

    # Step 2: Calculate the size of each category
    category_counts = adata.obs[strata_key].value_counts()

    # Step 3: Identify the top n largest categories or all categories if n is None
    if n_largest_groups is None:
        selected_categories = category_counts.index.tolist()
    else:
        selected_categories = category_counts.nlargest(n_largest_groups).index.tolist()

    # Step 4: Build adata_dict based on the strata key
    strata_dict = build_adata_dict(adata, [strata_key], desired_strata=selected_categories)

    # Step 5: Subsample each AnnData object in the strata_dict
    subsample_adata_dict(strata_dict, **kwargs)

    # Step 6: Drop AnnData objects with fewer than min_num_cells
    filtered_dict = {k: v for k, v in strata_dict.items() if v.n_obs >= min_num_cells}

    # Step 7: Concatenate the filtered_dict back to a single AnnData object
    #setting index_unique=None avoids index modification
    return concatenate_adata_dict(filtered_dict, index_unique=None)


def resample_adata_dict(
    adata_dict: AdataDict,
    strata_keys: list[str],
    n_largest_groups: int | None = None,
    min_num_cells: int = 0,
    **kwargs
) -> AdataDict:
    """
    Resample each :class:`AnnData` in ``adata_dict`` based on specified strata keys and drop strata with fewer than the minimum number of cells.

    Parameters
    -----------
    adata_dict
        An :class:`AdataDict`.

    strata_keys
        List of column names in `.obs` to use for stratification. Must be present in each adata in ``adata_dict``.

    min_num_cells
        Minimum number of cells required to retain a stratum.

    kwargs
        Additional keyword arguments to pass to :func:`~anndict.wrappers.resample_adata`.

    Returns
    --------
    :class:`AdataDict` of resampled :class:`AnnData` after filtering.
    """
    return adata_dict_fapply_return(adata_dict, resample_adata, strata_keys=strata_keys, n_largest_groups=n_largest_groups, min_num_cells=min_num_cells, **kwargs)


def normalize_adata_dict(
    adata_dict: AdataDict,
    **kwargs
) -> None:
    """
    Normalizes each :class:`AnnData` in ``adata_dict`` using :func:`sc.pp.normalize_total`.

    Parameters
    -----------
    adata_dict
        An :class:`AdataDict`.

    kwargs
        Additional keyword arguments to pass to :func:`sc.pp.normalize_total`

    Returns
    --------
    None

    Notes
    -----
    The function modifies ``adata_dict`` in-place.
    """
    adata_dict_fapply(adata_dict, sc.pp.normalize_total, **kwargs)


def log_transform_adata_dict(
    adata_dict: AdataDict,
    **kwargs
) -> None:
    """
    Log-transforms each :class:`AnnData` in ``adata_dict`` using :func:`sc.pp.log1p`.

    Parameters
    -----------
    adata_dict
        An :class:`AdataDict`.

    kwargs
        Additional keyword arguments to pass to :func:`sc.pp.log1p`

    Returns
    --------
    None

    Notes
    -----
    The function modifies ``adata_dict`` in-place.
    """
    adata_dict_fapply(adata_dict, sc.pp.log1p, **kwargs)


def set_high_variance_genes_adata_dict(
    adata_dict: AdataDict,
    **kwargs
) -> None:
    """
    Identifies highly variable genes in each :class:`AnnData` in ``adata_dict`` using :func:`sc.pp.highly_variable_genes`.

    Parameters
    -----------
    adata_dict
        An :class:`AdataDict`.

    kwargs
        Additional keyword arguments to pass to :func:`sc.pp.highly_variable_genes`

    Returns
    --------
    None

    Notes
    ------
    The function modifies ``adata_dict`` in-place.
    """
    adata_dict_fapply(adata_dict, sc.pp.highly_variable_genes, **kwargs)

def rank_genes_groups_adata_dict(
    adata_dict: AdataDict,
    **kwargs
) -> None:
    """
    Identifies differentially expressed genes in each :class:`AnnData` in ``adata_dict`` using :func:`sc.tl.rank_genes_groups`.

    Parameters
    -----------
    adata_dict
        An :class:`AdataDict`.

    kwargs
        Additional keyword arguments to pass to :func:`sc.tl.rank_genes_groups`

    Returns
    --------
    None

    Notes
    ------
    The function modifies ``adata_dict`` in-place.
    """
    adata_dict_fapply(adata_dict, sc.tl.rank_genes_groups, **kwargs)


def scale_adata_dict(
    adata_dict: AdataDict,
    **kwargs
) -> None:
    """
    Scales each :class:`AnnData` object in ``adata_dict`` using :func:`sc.pp.scale`.

    Parameters
    -----------
    adata_dict
        An :class:`AdataDict`.

    kwargs
        Additional keyword arguments to pass to :func:`sc.pp.scale`

    Returns
    --------
    None

    Notes
    -----
    The function modifies ``adata_dict`` in-place.
    """
    adata_dict_fapply(adata_dict, sc.pp.scale, **kwargs)


def pca_adata_dict(
    adata_dict: AdataDict,
    **kwargs
) -> None:
    """
    Performs PCA on each :class:`AnnData` object in ``adata_dict`` using :func:`sc.pp.pca`.

    Parameters
    -----------
    adata_dict
        An :class:`AdataDict`.

    kwargs
        Additional keyword arguments to pass to :func:`sc.pp.pca`.

    Returns
    --------
    None

    Notes
    -----
    The function modifies ``adata_dict`` in-place.
    """
    adata_dict_fapply(adata_dict, sc.pp.pca, **kwargs)


def neighbors_adata_dict(
    adata_dict: AdataDict,
    **kwargs
) -> None:
    """
    Calculates neighborhood graph for each :class:`AnnData` object in ``adata_dict`` using :func:`sc.pp.neighbors`.

    Parameters
    -----------
    adata_dict
        An :class:`AdataDict`.

    kwargs
        Additional keyword arguments to pass to :func:`sc.pp.neighbors`.

    Returns
    --------
    None

    Notes
    -----
    The function modifies ``adata_dict`` in-place.
    """
    adata_dict_fapply(adata_dict, sc.pp.neighbors, **kwargs)


def leiden_adata_dict(
    adata_dict: AdataDict,
    **kwargs
) -> None:
    """
    Performs Leiden clustering for each AnnData object in ``adata_dict`` using Scanpy's leiden function.

    Parameters
    -----------
    adata_dict
        An :class:`AdataDict`.

    kwargs
        Additional keyword arguments to pass to :func:`sc.tl.leiden`.

    Returns
    --------
    None

    Notes
    -----
    The function modifies ``adata_dict`` in-place.
    """
    adata_dict_fapply(adata_dict, sc.tl.leiden, **kwargs)

def leiden_sub_cluster(
    adata: AnnData,
    groupby: str,
    **kwargs
) -> AnnData:
    """
    Perform Leiden clustering on subgroups of cells.

    This function applies Leiden clustering to subgroups of cells defined by the ``groupby`` parameter.

    Parameters
    -----------
    adata
        An :class:`AnnData`.

    groupby
        Column name in ``adata.obs`` for grouping cells before subclustering.

    kwargs
        Additional keyword arguments to pass to :func:`leiden_adata_dict`.

    Returns
    --------
    None

    Notes
    -----
    The function modifies the input ``AnnData`` object in-place.
    """
    adata_dict = build_adata_dict(adata, strata_keys=[groupby])
    leiden_adata_dict(adata_dict, **kwargs)
    adata = concatenate_adata_dict(adata_dict, index_unique=None) #setting index_unique=None avoids index modification
    return adata


def leiden_sub_cluster_adata_dict(
    adata_dict: AdataDict,
    groupby: str,
    **kwargs
) -> AdataDict:
    """
    Applies the :func:`leiden_sub_cluster` function to each :class:`AnnData` object in ``adata_dict``.

    Parameters
    -----------
    adata_dict
        An :class:`AdataDict`.

    groupby
        Column name in ``adata.obs`` for grouping cells before subclustering.

    kwargs
        Additional keyword arguments to pass to :func:`leiden_sub_cluster`.

    Returns
    --------
    :class:`AdataDict` of resampled AnnData objects after filtering.
    """
    return adata_dict_fapply_return(adata_dict, leiden_sub_cluster, groupby=groupby, **kwargs)


def calculate_umap_adata_dict(
    adata_dict: AdataDict,
    **kwargs
) -> AdataDict:
    """
    Calculates UMAP embeddings for each :class:`AnnData` in ``adata_dict`` using :func:`sc.tl.umap`.

    Parameters
    -----------
    adata_dict
        An :class:`AdataDict`.

    kwargs
        Additional keyword arguments to pass to :func:`sc.tl.umap`.

    Returns
    --------
    :class:`AdataDict` with UMAP coordinates added to each :class:`AnnData`.
    """
    adata_dict_fapply(adata_dict, sc.tl.umap, **kwargs)
    return adata_dict


def plot_umap_adata_dict(
    adata_dict: AdataDict,
    **kwargs
) -> None:
    """
    Plots UMAP embeddings for each :class:`AnnData` in ``adata_dict``, colored by a specified variable.

    Parameters
    -----------
    adata_dict
        An :class:`AdataDict`.

    kwargs
        Additional keyword arguments, including 'color_by' which specifies a variable by which to color the UMAP plots, typically a column in ``.obs``.

    Returns
    --------
    None

    Notes
    -----
    The function creates plots for the each :class:`AnnData`.
    """
    def plot_umap(adata, adt_key=None, **kwargs):
        print(f"Plotting UMAP for key: {adt_key}")
        if 'X_umap' in adata.obsm:
            sc.pl.umap(adata, **kwargs)
        else:
            print(f"UMAP not computed for adata with key {adt_key}. Please compute UMAP before plotting.")
    adata_dict_fapply(adata_dict, plot_umap, use_multithreading=False, **kwargs)


def plot_spatial_adata_dict(
    adata_dict: AdataDict,
     **kwargs
) -> None:
    """
    Plots spatial data for each :class:`AnnData` in ``adata_dict``, colored by a specified variable.

    Parameters
    ----------
    adata_dict
        An :class:`AdataDict`.

    kwargs
        Additional keyword arguments, including 'color_by' which specifies a variable by which to color the spatial plots, typically a column in .obs, and 'crop_coord' which specifies coordinates for cropping the spatial plots.

    Returns
    -------
    None

    Notes
    ------
    The function creates spatial plots for each :class:`AnnData` in ``adata_dict``.
    """
    def plot_spatial(adata, **kwargs):
        if 'spatial' in adata.obsm:
            sc.pl.spatial(adata, **kwargs)
        else:
            print("Spatial coordinates not available for adata. Please add spatial data before plotting.")

    adata_dict_fapply(adata_dict, plot_spatial, use_multithreading=False, **kwargs)


"""
This module contains the functions write AdataDict objects to disk.
"""

import os
import json
import scanpy as sc

from anndata import AnnData

from .adata_dict import AdataDict

def write_adata_dict(
    adata_dict: AdataDict,
    directory: str,
    *,
    file_prefix: str = "",
) -> None:
    """
    Save each :class:`AnnData` object from an :class:`AdataDict` into a separate ``.h5ad`` file,
    creating a directory structure that reflects the hierarchy of the :class:`AdataDict`
    using key values as directory names. The hierarchy is saved in a file called
    ``adata_dict.hierarchy`` in the top-level directory.

    Parameters
    ----------
    adata_dict
        An :class:`AdataDict`.

    directory
        Base directory where ``.h5ad`` files will be saved.

    file_prefix
        Optional prefix for the filenames.

    Notes
    -----
    Each element in the key tuple becomes a subdirectory, ensuring the directory hierarchy 
    matches the AdataDict's hierarchy precisely. The filename itself is ``<file_prefix> + 
    key_elements_joined_by_underscores.h5ad``.

    Examples
    --------
    **Case 1: Flat hierarchy**

    .. code-block:: python

        adata_dict.set_hierarchy(["Donor", "Tissue"])
        print(adata_dict)
        > {
        >     ("Donor1", "Tissue1"): adata_d1_t1,
        >     ("Donor1", "Tissue2"): adata_d1_t2,
        >     ("Donor2", "Tissue1"): adata_d2_t1,
        > }

    The files will be saved as:

    .. code-block:: bash

        directory/
        ├── adata_dict.hierarchy.json
        ├── adata_dict.db.json
        ├── <file_prefix>Donor1_Tissue1.h5ad
        ├── <file_prefix>Donor1_Tissue2.h5ad
        └── <file_prefix>Donor2_Tissue1.h5ad

    **Case 2: Nested hierarchy**

    .. code-block:: python

        adata_dict.set_hierarchy(["Donor", ["Tissue"]])
        print(adata_dict)
        > {
        >     ("Donor1",): {
        >         ("Tissue1",): adata_d1_t1,
        >         ("Tissue2",): adata_d1_t2,
        >     },
        >     ("Donor2",): {
        >         ("Tissue1",): adata_d2_t1,
        >     },
        > }

    The files will be saved as:

    .. code-block:: bash

        directory/
        ├── adata_dict.hierarchy.json
        ├── adata_dict.db.json
        ├── Donor1/
        │   ├── <file_prefix>Tissue1.h5ad
        │   └── <file_prefix>Tissue2.h5ad
        └── Donor2/
            └── <file_prefix>Tissue1.h5ad

    **Case 3: Nested hierarchy with multiple indices at the deepest level**

    .. code-block:: python

        adata_dict.set_hierarchy(["Donor", ["Tissue", "Cell Type"]])
        print(adata_dict)
        > {
        >     ("Donor1",): {
        >         ("Tissue1", "CellType1"): adata_d1_t1_c1,
        >         ("Tissue1", "CellType2"): adata_d1_t1_c2,
        >         ("Tissue2", "CellType3"): adata_d1_t2_c3,
        >     },
        >     ("Donor2",): {
        >         ("Tissue1", "CellType1"): adata_d2_t1_c1,
        >     },
        > }

    The files will be saved as:

    .. code-block:: bash

        directory/
        ├── adata_dict.hierarchy.json
        ├── adata_dict.db.json
        ├── Donor1/
        │   ├── <file_prefix>Tissue1_CellType1.h5ad
        │   ├── <file_prefix>Tissue1_CellType2.h5ad
        │   └── <file_prefix>Tissue2_CellType3.h5ad
        └── Donor2/
            └── <file_prefix>Tissue1_CellType1.h5ad
    """

    os.makedirs(directory, exist_ok=False)

    # Save the original hierarchy for complete reconstruction
    hierarchy_file_path = os.path.join(directory, "adata_dict.hierarchy.json")
    with open(hierarchy_file_path, "w", encoding="utf-8") as f:
        json.dump(adata_dict.hierarchy, f)

    db_file_path = os.path.join(directory, "adata_dict.db.json")
    db_entries = []

    def safe_str(val):
        s = str(val)
        return "".join(c if c not in r'\/:*?"<>|' else "_" for c in s)

    # Recursively traverse the AdataDict, yielding:
    #   full_key: tuple of all key elements from outer levels plus current level.
    #   file_key: tuple of key elements at the deepest level (used for file naming).
    #   value: the AnnData object.
    def iter_items(d, parent_key=()):
        for key_tuple, value in d.items():
            full_key = parent_key + key_tuple
            if isinstance(value, dict):
                yield from iter_items(value, full_key)
            elif isinstance(value, AnnData):
                yield full_key, key_tuple, value
            else:
                raise ValueError("Unexpected value type in adata_dict; expected AnnData or dict.")

    # Process each AnnData in the nested structure.
    for full_key, file_key, adata in iter_items(adata_dict):
        # The directory path is built from the outer (non-deepest) keys.
        dir_keys = full_key[:-len(file_key)]
        sub_dir = os.path.join(directory, *[safe_str(x) for x in dir_keys]) if dir_keys else directory
        os.makedirs(sub_dir, exist_ok=True)
        # The file name is built solely from the deepest-level key tuple.
        file_name = f"{file_prefix}{'_'.join(safe_str(x) for x in file_key)}.h5ad"
        file_path = os.path.join(sub_dir, file_name)
        sc.write(file_path, adata)
        rel_file_path = os.path.relpath(file_path, start=directory)
        db_entries.append({"file_path": rel_file_path, "key": list(full_key)})

    with open(db_file_path, "w", encoding="utf-8") as f:
        json.dump(db_entries, f)



"""
Clean single columns in ``adata.var``
"""

import ast

import pandas as pd

from anndata import AnnData

from anndict.utils import enforce_semantic_list
from anndict.llm import (
    retry_call_llm,
    extract_dictionary_from_ai_string,
    process_llm_category_mapping
)

def simplify_var_index(
    adata: AnnData,
    column: str,
    new_column_name: str,
    simplification_level: str = ''
    ) -> dict:
    """

    Simplifies gene names in the index of the :class:`AnnData` object's var 
    attribute based on a boolean column, and stores the result in a new 
    column using the :func:`map_gene_labels_to_simplified_set`. This function 
    assumes that ``adata.var`` contains gene symbols (i.e. PER1, IL1A) 
    and not numeric indices or accession numbers.

    Parameters
    ------------
    adata
        The :class:`AnnData` object containing the data.

    column
        The boolean column in ``adata.var`` used to select genes for simplification.

    new_column_name
        The name of the new column to store the simplified labels.

    simplification_level
        A qualitative description of how much you want the labels to be simplified. 
        Could be anything, like ``'extremely'``, ``'barely'``, or ``'pathway-level'``.

    Returns
    --------
    A :class:`dict` containing the map from the current labels to the simplified labels

    Raises
    --------
    ValueError
        If more than 1000 genes are selected for simplification or if the 
        masking column (used to select genes) is not boolean.

    Notes
    -------
    Modifies ``adata`` by adding ``adata.var[new_column_name]`` (i.e. the new labels) in-place.

    Example
    ---------
    .. code-block:: python

        import anndict as adt

        print(adata.var)
        >  index        simplify
        > 'HSP90AA1'    1
        > 'HSPA1A'      1
        > 'HSPA1B'      1
        > 'CLOCK'       1
        > 'ARNTL'       1
        > 'PER1'        1
        > 'IL1A'        1
        > 'IL6'         1
        > 'APOD'        0
        > 'CFD'         0

        label_mapping = adt.simplify_var_index(adata,
                                '',
                                new_column_name = 'functional_category',
                                simplification_level='functional category level'
                                )

        print(adata.var) # New column added
        >  index        simplify        functional_category
        > 'HSP90AA1'    1               'Heat Shock Proteins'
        > 'HSPA1A'      1               'Heat Shock Proteins'
        > 'HSPA1B'      1               'Heat Shock Proteins'
        > 'CLOCK'       1               'Circadian Rythm'
        > 'ARNTL'       1               'Circadian Rythm'
        > 'PER1'        1               'Circadian Rythm'
        > 'IL1A'        1               'Interleukin'
        > 'IL6'         1               'Interleukin'
        > 'APOD'        0                Nan
        > 'CFD'         0                Nan

    """
    if not pd.api.types.is_bool_dtype(adata.var[column]):
        raise ValueError(f"The column '{column}' must be a boolean column.")

    # Get the indices of True in the boolean column
    selected_genes = adata.var.index[adata.var[column]]

    if len(selected_genes) > 1000:
        raise ValueError("Cannot simplify more than 1000 genes at a time.")

    # Get the mapping of original labels to simplified labels using the provided function
    label_mapping = map_gene_labels_to_simplified_set(
        selected_genes, simplification_level=simplification_level)

    # Apply the mapping to create the new column in the AnnData object
    adata.var[new_column_name] = adata.var.index.to_series()\
        .map(label_mapping).fillna(adata.var.index.to_series())

    return label_mapping


def map_gene_labels_to_simplified_set(
    labels: list[str],
    simplification_level: str = '',
    batch_size: int = 50
    ) -> dict:
    """
    Maps a list of genes to a simplified set of labels using an LLM, processing in batches.

    Parameters
    -----------
    labels
        The list of labels to be mapped.

    simplification_level
        A qualitative description of how much you want the labels to be simplified.

    batch_size
        The number of labels to process in each batch.

    Returns
    ----------
    A :class:`dict` mapping the original labels to the simplified set of labels.

    Example
    ---------
    .. code-block:: python

        import anndict as adt

        gene_labels = ['HSP90AA1',
                       'HSPA1A',
                       'HSPA1B',
                       'CLOCK',
                       'ARNTL',
                       'PER1',
                       'IL1A',
                       'IL6'
                       ]

        label_mapping = adt.map_gene_labels_to_simplified_set(gene_labels,
                                simplification_level='functional category level'
                                )

        print(label_mapping)

        > {
        >     'HSP90AA1': 'Heat Shock Protein',
        >     'HSPA1A': 'Heat Shock Protein',
        >     'HSPA1B': 'Heat Shock Protein',
        >     'CLOCK': 'Circadian Rhythm',
        >     'ARNTL': 'Circadian Rhythm',
        >     'PER1': 'Circadian Rhythm',
        >     'IL1A': 'Interleukin',
        >     'IL6': 'Interleukin'
        > }
    """
    # Enforce that labels are semantic
    enforce_semantic_list(labels)

    # Prepare the initial prompt
    initial_labels_str = "    ".join(labels)

    # Prepare the messages for the Chat Completions API
    messages = [
        {"role": "system",
        "content": f"You are a python dictionary mapping generator that takes \
                a list of genes and provides a mapping to a {simplification_level} \
                simplified set as a dictionary. \
                Example: HSP90AA1    HSPA1A    HSPA1B    CLOCK    ARNTL    PER1    IL1A    IL6 -> \
                {{'HSP90AA1':'Heat Shock Proteins','HSPA1A':'Heat Shock Proteins',\
                'HSPA1B':'Heat Shock Proteins','CLOCK':'Circadian Rhythm',\
                'ARNTL':'Circadian Rhythm','PER1':'Circadian Rhythm',\
                'IL1A':'Interleukins','IL6':'Interleukins'}}"},
        {"role": "user",
        "content": f"Here is the full list of gene labels to be simplified: \
                {initial_labels_str}. Acknowledge that you've seen all labels. \
                Do not provide the mapping yet."}
    ]

    # Get initial acknowledgment
    initial_response = retry_call_llm(
        messages=messages,
        process_response=lambda x: x,
        failure_handler=lambda: "Failed to process initial prompt",
        call_llm_kwargs={'max_tokens': 30, 'temperature': 0},
        max_attempts=1
    )
    messages.append({"role": "assistant", "content": initial_response})

    def process_batch(batch_labels):
        batch_str = "    ".join(batch_labels)
        messages.append({"role": "user", \
                            "content": f"Provide a mapping for this batch of gene labels. \
                            Generate only a dictionary: {batch_str} -> "})

        def process_response(response):
            cleaned_mapping = extract_dictionary_from_ai_string(response)
            return ast.literal_eval(cleaned_mapping)

        def failure_handler(labels):
            print(f"Simplification failed for gene labels: {labels}")
            return {label: label for label in labels}

        call_llm_kwargs = {
            'max_tokens': min(300 + 25*len(batch_labels), 4000),
            'temperature': 0
        }
        failure_handler_kwargs = {'labels': batch_labels}

        batch_mapping = retry_call_llm(
            messages=messages,
            process_response=process_response,
            failure_handler=failure_handler,
            call_llm_kwargs=call_llm_kwargs,
            failure_handler_kwargs=failure_handler_kwargs
        )
        messages.append({"role": "assistant", "content": str(batch_mapping)})
        return batch_mapping

    # Process all labels in batches
    full_mapping = {}
    for i in range(0, len(labels), batch_size):
        batch = labels[i:i+batch_size]
        batch_mapping = process_batch(batch)
        full_mapping.update(batch_mapping)

    # Final pass to ensure consistency
    final_mapping = process_llm_category_mapping(labels, full_mapping)

    return final_mapping



"""
This subpackage contains functions to annotate and assess cell annotation.
"""

#annotate cells de novo (from scratch)
from .de_novo import (
    ai_annotate, ai_annotate_by_comparison,
    ai_cell_type, ai_annotate_cell_type,
    ai_cell_types_by_comparison,
    ai_annotate_cell_type_by_comparison,
    ai_annotate_cell_type_by_comparison_adata_dict,
    ai_annotate_cell_sub_type,
    ai_annotate_biological_process,
    cell_type_marker_gene_score,
    ai_determine_leiden_resolution
)

#transfer labels from an already labeled adata to an unlabaled adata
from .label_transfer import (
    harmony_label_transfer,
    stable_label, stable_label_adata
)

#analyze annotation results
from .benchmarking import (
    create_label_df,
    kappa_adata, krippendorff_alpha_adata,
    ai_compare_cell_type_labels_pairwise
)


__all__ = [
    # De Novo
    #--------
    # base.py
    "ai_annotate",
    "ai_annotate_by_comparison",

    # annotate_cell_type.py
    "ai_cell_type",
    "ai_annotate_cell_type",

    # annotate_cell_type_by_comparison.py
    "ai_cell_types_by_comparison",
    "ai_annotate_cell_type_by_comparison",

    # annotate_cell_subtype.py
    "ai_annotate_cell_type_by_comparison_adata_dict",
    "ai_annotate_cell_sub_type",

    # annotate_biological_process.py
    "ai_annotate_biological_process",

    # cell_type_score.py
    "cell_type_marker_gene_score",

    # automated_clustering.py
    "ai_determine_leiden_resolution",

    # Label Transfer
    #--------------
    # harmony.py
    "harmony_label_transfer",

    # stabilizing_classifier.py
    "stable_label",
    "stable_label_adata",

    # Benchmarking
    #------------
    # label_utils.py
    "create_label_df",

    # agreement_metrics.py
    "kappa_adata",
    "krippendorff_alpha_adata",

    # ai_compare_cell_type_labels.py
    "ai_compare_cell_type_labels_pairwise",
]


"""
utils for :class:`AdataDict`
"""

from __future__ import annotations  # allows type hinting without circular dependency
from typing import TYPE_CHECKING

from .adata_dict_fapply import adata_dict_fapply_return

if TYPE_CHECKING:
    from .adata_dict import AdataDict


def to_nested_tuple(nested_list):
    """
    Recursively convert a nested list into a nested tuple.
    """
    if isinstance(nested_list, list):
        return tuple(to_nested_tuple(item) for item in nested_list)
    return nested_list


def to_nested_list(nested_tuple):
    """
    Recursively convert a nested tuple into a nested list.
    """
    if isinstance(nested_tuple, tuple):
        return list(to_nested_list(item) for item in nested_tuple)
    return nested_tuple


def set_var_index_func(adata_dict: AdataDict, column: str) -> AdataDict:
    """
    Set the index of ``adata.var`` to the specified
    column for each :class:`AnnData` in ``adata_dict``.

    Parameters
    ------------
    adata_dict
        An :class:`AdataDict`.

    column
        The column name to set as the index of ``adata.var``.

    Returns
    --------
    An :class:`AdataDict` with the same structure as
    ``adata_dict``, where the ``.var`` attribute of each :class:`AnnData`
    object has its index set to the specified column.
    """

    def set_var_index_main(adata, column):
        adata.var = adata.var.set_index(column)
        return adata

    return adata_dict_fapply_return(
        adata_dict, set_var_index_main, column=column, return_as_adata_dict=True
    )


def set_obs_index_func(adata_dict: AdataDict, column: str) -> AdataDict:
    """
    Set the index of adata.obs to the specified column for each AnnData object in adata_dict.

    Parameters
    -----------
    adata_dict
        An :class:`AdataDict`.

    column
        The column name to set as the index of ``adata.obs``.

    Returns
    --------
    An :class:`AdataDict` with the same structure as 
    ``adata_dict``, where the ``.obs`` attribute of 
    each :class:`AnnData` object has its index set to the specified column.
    """

    def set_obs_index_main(adata, column):
        adata.obs = adata.obs.set_index(column)
        return adata

    return adata_dict_fapply_return(
        adata_dict, set_obs_index_main, column=column, return_as_adata_dict=True
    )



"""
This module handles annotating a set of genes sets by considering each gene set in the context of the other sets of genes.
"""

from anndata import AnnData

from anndict.utils import enforce_semantic_list
from anndict.llm import retry_call_llm
from .base import ai_annotate_by_comparison


def ai_cell_types_by_comparison(
    gene_lists: list[list[str]],
    cell_types: list[str] = None,
    tissues: list[str] = None,
    subtype: bool = False
) -> list[str]:
    """
    Returns cell type labels for multiple lists of marker genes as determined by an LLM.

    Parameters
    -------------
    gene_lists
        A list containing multiple lists of genes to be described.

    cell_types
        A list of cell types to provide context to the LLM. 
        Each entry corresponds to a gene list in ``gene_lists``.

    tissues
        A list of tissues of origin to provide context to the LLM. 
        Each entry corresponds to a gene list in ``gene_lists``.

    subtype
        Whether to try to annotate cell type (if ``False``) or subtype (if ``True``).

    Returns
    --------
    The cell type labels generated by an LLM for each gene list.
    """
    if not gene_lists:
        raise ValueError("Empty list of gene lists passed to ai_cell_type_by_comparison")

    # Enforce semantic_list for each gene list
    for gene_list in gene_lists:
        enforce_semantic_list(gene_list)

    # Prepare the system prompt
    system_prompt = (
        "You are a terse molecular biologist. You respond in a few words \
        and without restating any part of the question. "
        f"Compare and contrast gene sets to identify the most likely \
        cell {'sub' if subtype else ''}type based on marker genes."
    )

    # Prepare the initial user prompt for contrasting all gene lists
    # initial_prompt = f"Tissue: {tissues}, " if tissue else ""
    # initial_prompt += f"Cell Type: {cell_type}, " if cell_type else ""
    initial_prompt = "Briefly compare and contrast the following gene sets:\n"
    for i, gene_list in enumerate(gene_lists, 1):
        tissue_str = " " + ', '.join(tissues[i]) if tissues and tissues[i] else ""
        cell_type_str = " " + ', '.join(cell_types[i]) if cell_types and cell_types[i] else ""

        initial_prompt += f"{i}){tissue_str}{cell_type_str} {('    '.join(gene_list))}\n"

    # Initialize the conversation
    messages = [
        {"role": "system", "content": system_prompt},
        {"role": "user", "content": initial_prompt}
    ]

    # Get the initial contrast response
    contrast_response = retry_call_llm(
        messages=messages,
        process_response=lambda x: x,
        failure_handler=lambda: "Failed to contrast gene sets",
        call_llm_kwargs={'max_tokens': 300, 'temperature': 0},
        max_attempts=1
    )

    # Append the contrast response to the conversation
    messages.append({"role": "assistant", "content": contrast_response})

    messages.append({"role": "user", "content": "Provide only the new label. "})

    # Process each gene list
    cell_subtype_labels = []
    for i, gene_list in enumerate(gene_lists, 1):
        tissue_str = " " + ', '.join(tissues[i]) if tissues and tissues[i] else ""
        cell_type_str = " " + ', '.join(cell_types[i]) if cell_types and cell_types[i] else ""

        gene_set_prompt = f"What is the cell{tissue_str}{cell_type_str} {'sub' if subtype else ''}type \
            label for the gene set: {('    '.join(gene_list))}?"
        messages.append({"role": "user", "content": gene_set_prompt})

        # Get the subtype label
        subtype_label = retry_call_llm(
            messages=messages,
            process_response=lambda x: x.strip(),
            failure_handler=lambda cell_type_str=cell_type_str, i=i: cell_type_str if cell_types and cell_types[i] else "Unknown",
            call_llm_kwargs={'max_tokens': 50, 'temperature': 0},
            max_attempts=1
        )

        cell_subtype_labels.append(subtype_label)
        messages.append({"role": "assistant", "content": subtype_label})

    # print(f"{messages}")
    return cell_subtype_labels


def ai_annotate_cell_type_by_comparison(
    adata: AnnData,
    groupby: str,
    n_top_genes: int,
    *,
    new_label_column: str ='ai_cell_type_by_comparison',
    cell_type_of_origin_col: str = None,
    tissue_of_origin_col: str = None,
    **kwargs
) -> AnnData:
    """
    Annotate cell types by comparison using LLMs.

    This function wraps the ai_annotate function to perform cell type annotation
    based on comparison of gene expression profiles.

    Parameters
    -------------
    adata
        AnnData Annotated data matrix.

    groupby
        Column name in adata.obs for grouping cells.

    n_top_genes
        Number of top genes to consider for annotation.

    new_label_column
        Name of the column to store the AI-generated cell type labels.

    cell_type_of_origin_col
        Name of a column in ``adata.obs`` that contains the cell type 
        of orgin. Used for context to the LLM.

    tissue_of_origin_col
        Name of a column in ``adata.obs`` that contains the tissue of orgin. 
        Used to provide context to the LLM.

    adt_key
        Used by :func:`adata_dict_fapply` or :func:`adata_dict_fapply_return` 
        when passing this function to them.

    Returns
    -------
    :class:`AnnData` with LLM-generated cell type labels stored in ``adata.obs[label_column]``.
    """
    return ai_annotate_by_comparison(func=ai_cell_types_by_comparison, adata=adata, groupby=groupby, n_top_genes=n_top_genes, new_label_column=new_label_column, cell_type_of_origin_col=cell_type_of_origin_col, tissue_of_origin_col=tissue_of_origin_col, **kwargs)



"""
Filters cells based on their density in PCA space.
Can be useful to remove outliers, missannotated cells, or to get a more homogenous population.
"""

from warnings import warn

import numpy as np

from sklearn.decomposition import PCA
from scipy.stats import gaussian_kde
from scipy import sparse
from scipy.sparse import spmatrix
from anndata import AnnData

from anndict.utils.anndata_ import add_col_to_adata_obs, add_col_to_adata_var


def pca_density_filter_adata(
    adata: AnnData,
    use_layer: str | None = None,
    new_filter_column_name: str = "density_filter",
    drop_cells: bool = False,
    **kwargs
):
    """
    Filters  :func:`pca_density_filter` that accepts :class:`AnnData` as input.

    Parameters
    -----------
    adata
        An :class:`AnnData`.

    use_layer
        name of the layer to use for PCA (default: None, use adata.X).

    new_filter_column_name
        Name of the new boolean column to be created in ``adata.obs``.

    **kwargs
        Additional keyword arguments to pass to :func:`pca_density_filter_main`.

    Returns
    --------
    If ``drop_cells`` is ``True``, returns a new :class:`AnnData` with the filtered cells.
    Otherwise, modifies ``adata`` in-place.
    """
    # Get the data to use for PCA
    data = adata.layers[use_layer] if use_layer in adata.layers else adata.X

    # Calculate density (skip if fewer than 10 cells)
    if data.shape[0] < 10:
        density, cutoff, variables_used_in_pca = (
            np.ones(data.shape[0]),
            0,
            np.zeros(data.shape[1], dtype=bool),
        )
    else:
        density, cutoff, variables_used_in_pca = pca_density_filter_main(data, **kwargs)

    # Mark cells above the threshold
    high_density_indices = density > cutoff
    cell_passed_filter_bool = np.zeros(data.shape[0], dtype=bool)
    cell_passed_filter_bool[high_density_indices] = True

    # record which cells were passed
    # Note: this function converts index_vector to
    # ``pd.BooleanDtype`` (from ``bool``) when adding col to adata
    add_col_to_adata_obs(
        adata, np.arange(data.shape[0]), cell_passed_filter_bool, new_filter_column_name
    )

    # record which variables were used in PCA
    # Note: this function converts variables_used_in_pca to
    # ``pd.BooleanDtype`` (from ``bool``) when adding col to adata
    variable_used_in_pca_bool = np.zeros(adata.var.shape[0], dtype=bool)
    variable_used_in_pca_bool[variables_used_in_pca] = True
    add_col_to_adata_var(
        adata,
        np.arange(adata.var.shape[0]),
        variable_used_in_pca_bool,
        new_filter_column_name,
    )

    # Drop cells if requested
    if drop_cells:
        adata = adata[cell_passed_filter_bool].copy()
        return adata


def pca_density_filter_main(
    data: np.ndarray | spmatrix, n_components: int = 3, threshold: float = 0.10
) -> tuple[np.ndarray, float, list[int]]:
    """
    Calculate density contours for PCA-reduced data, return the density of all input data,
    and identify the indices of the variables that were included in the PCA.

    Parameters
    -----------
    data
        array-like, shape (n_samples, n_features)

    n_components
        number of components for PCA to reduce the data to.

    threshold
        percentile threshold for density values. Points with
        density above this threshold are considered high-density and will be retained.

    Returns
    --------
    A tuple containing:
    - pca_data: PCA-reduced data (None if all variables are constant).
    - density: Density values of all the points (None if all variables are constant).
    - variables_used_in_pca: Indices of unique variables that were included in
        the PCA (empty list if all variables are constant).
    """

    # Convert to dense if sparse
    if sparse.issparse(data):
        data = data.toarray()

    # Check for constant variables (these will not be used by PCA)
    non_constant_columns = np.var(data, axis=0) > 0
    variables_used_in_pca = np.arange(data.shape[1])[non_constant_columns]

    # Skip the block if no non-constant variables are found
    if not np.any(non_constant_columns):
        return None, None, []

    # Adjust n_components if necessary
    n_features = np.sum(non_constant_columns)
    n_samples = data.shape[0]
    n_components = min(n_components, n_features, n_samples)

    # Perform PCA reduction only on non-constant variables
    pca = PCA(n_components=n_components)
    pca_data = pca.fit_transform(data[:, non_constant_columns])

    # Calculate the point density for all points
    kde = gaussian_kde(pca_data.T)
    density = kde(pca_data.T)

    # Determine the density threshold
    cutoff = np.percentile(density, threshold * 100)

    return density, cutoff, variables_used_in_pca.tolist()


def pca_density_subsets(data: np.ndarray | spmatrix, labels: np.ndarray) -> np.ndarray:
    """
    Apply calculate_density_contours_with_unique_variables to subsets of X indicated by labels.
    Returns a vector indicating whether each row in X is above the threshold for its respective
    label group.

    Parameters
    -----------
    data
        array-like, shape (n_samples, n_features)
    labels
        array-like, shape (n_samples,), labels indicating the subset to which each row belongs

    Returns
    --------
    index_vector
        array-like, boolean vector of length n_samples indicating rows above the threshold
    """
    warn(
        "pca_density_subsets() is deprecated and will be removed \
        in a future version. Use pca_density_filter_adata() or pca_density_filter_main() instead.",
        DeprecationWarning,
        stacklevel=2,
    )

    unique_labels = np.unique(labels)
    index_vector = np.zeros(len(data), dtype=bool)

    for label in unique_labels:
        subset = data[labels == label]
        if subset.shape[0] < 10:
            # If fewer than 10 cells, include all cells by assigning density = 1 and cutoff = 0
            density, cutoff = np.ones(subset.shape[0]), 0
        else:
            density, cutoff, _ = pca_density_filter_main(
                subset, n_components=3, threshold=0.10
            )

        # Mark rows above the threshold for this label
        high_density_indices = density > cutoff
        global_indices = np.where(labels == label)[0][high_density_indices]
        index_vector[global_indices] = True

    return index_vector



"""
Make (grouped) barplots of agreement scores between columns in ``adata.obs``.
"""


import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

from matplotlib.figure import Figure
from matplotlib.axes import Axes
from seaborn.matrix import ClusterGrid
from seaborn import FacetGrid
from anndata import AnnData


def plot_grouped_average(
    adata: AnnData,
    label_value: dict[str,str],
    adt_key: tuple[str,...] | None = None
):
    """
    Plots the average values specified in label_value across each group of label_keys in an AnnData object.

    Parameters
    -----------
    adata: 
        An :class:`AnnData`.

    label_value: 
        A :class:`dict` where the keys are the cols in ``adata.obs`` for grouping, values are the cols in ``adata.obs`` for the values to average.

    adt_key: 
        to print specified key

    Returns
    --------
    None
    """
    print(adt_key)
    if not all(label in adata.obs for label in label_value.keys()):
        missing_keys = [label for label in label_value.keys() if label not in adata.obs]
        raise ValueError(f"Label key(s) {missing_keys} not found in adata.obs.")
    if not all(value in adata.obs for value in label_value.values()):
        missing_values = [value for value in label_value.values() if value not in adata.obs]
        raise ValueError(f"Value key(s) {missing_values} not found in adata.obs.")

    grouped_means = {}
    for label, value in label_value.items():
        grouped_means[label] = adata.obs.groupby(label)[value].mean()

    # Create a DataFrame from the grouped means
    df = pd.DataFrame(grouped_means)

    # Plot the results
    df.plot(kind='bar', figsize=(12, 8), color=plt.cm.get_cmap("Paired").colors)
    plt.xlabel('Groups')
    plt.ylabel('Average Scores')
    plt.title('Average Scores across Groups')
    plt.xticks(rotation=90)
    plt.legend(title='Scores')
    plt.show()


def plot_model_agreement(
    adata: AnnData,
    group_by: str,
    sub_group_by: str,
    agreement_cols: list[str],
    granularity: int = 2
) -> tuple[Figure, Axes] | ClusterGrid:
    """
    Plots the average values of specified agreement columns across varying levels of granularity.

    Parameters
    -----------
    adata
        an :class:`AnnData`.

    group_by
        Column name in ``adata.obs`` for the main grouping (e.g., ``'cell_type'``).

    sub_group_by
        Column name in ``adata.obs`` for the sub-grouping (e.g., ``'tissue'``).

    agreement_cols
        Column names specifying the agreement columns (e.g., ``['agreement_of_manual_with_model1', 'agreement_of_manual_with_model2']``).

    granularity
        Level of detail in the plot (``0`` = models only, ``1`` = models within cell types, ``2`` = models within cell types and tissues).

    Returns
    --------
    The plot of agreement, averaged according to the ``granularity`` setting.

    """
    if not all(col in adata.obs for col in agreement_cols):
        missing_cols = [col for col in agreement_cols if col not in adata.obs]
        raise ValueError(f"Columns {missing_cols} not found in adata.obs.")
    if group_by not in adata.obs:
        raise ValueError(f"Group key '{group_by}' not found in adata.obs.")
    if sub_group_by not in adata.obs:
        raise ValueError(f"Sub-group key '{sub_group_by}' not found in adata.obs.")

    # Pivot longer to get columns: group_by, sub_group_by, agreement, model_name
    melted = adata.obs.melt(id_vars=[group_by, sub_group_by], value_vars=agreement_cols,
                            var_name='model_name', value_name='agreement')

    if granularity == 0:
        # Calculate the average scores across all groups within each model
        grouped_means = melted.groupby('model_name')['agreement'].mean().sort_values(ascending=False)

        # Create figure and axis objects
        fig, ax = plt.subplots(figsize=(14, 8))

        # Plot the bar chart
        grouped_means.plot(kind='bar', ax=ax, colormap='Paired')

        # Add value labels on top of each bar
        for i, v in enumerate(grouped_means):
            ax.text(i, v, f'{v * 100:.0f}%', ha='center', va='bottom')

    elif granularity == 1:
        # Calculate the average scores within each model and cell type
        grouped_means = melted.groupby([group_by, 'model_name'])['agreement'].mean().unstack()

        fig, ax = plt.subplots(figsize=(14, 8))
        grouped_means.plot(kind='bar', ax=ax, colormap='Paired')

    elif granularity == 2:
        # Calculate the average scores within each model, cell type, and tissue
        grouped_means = melted.groupby([group_by, sub_group_by, 'model_name'])['agreement'].mean().unstack(level=[1, 2])

        # Ensure the data is numeric and allow NaNs (missing values)
        grouped_means = grouped_means.apply(pd.to_numeric, errors='coerce')

        # Create a mask for NaN values
        mask = grouped_means.isnull()

        # Create a color mapping for tissues using the provided colors
        tissue_colors = [
            "#1f77b4", "#aec7e8", "#ff7f0e", "#ffbb78",
            "#2ca02c", "#98df8a", "#d62728", "#ff9896",
            "#9467bd", "#c5b0d5", "#8c564b", "#c49c94",
            "#e377c2", "#f7b6d2", "#7f7f7f", "#c7c7c7",
            "#bcbd22", "#dbdb8d", "#17becf", "#9edae5",
            "#7f9ec0", "#ffab60", "#5ab4ac",
            # 5 additional concordant colors
            "#8b4513", "#ff6347", "#4682b4", "#dda0dd", "#ffd700"
        ]

        # Ensure that the number of tissues does not exceed the number of available colors
        tissues = grouped_means.columns.get_level_values(0).unique()
        tissue_colors = tissue_colors[:len(tissues)]

        # Create a color map based on the provided colors
        tissue_color_map = dict(zip(tissues, tissue_colors))

        # Create column colors based on tissues
        col_colors = [tissue_color_map[tissue] for tissue in grouped_means.columns.get_level_values(0)]

        # Plot heatmap with col_colors
        # fig, ax = plt.subplots(figsize=(16, 10))
        # Create the clustermapimport seaborn as sns

        # Use the 'viridis_r' colormap
        cmap = plt.get_cmap('viridis_r')

        # Set the color for NaN values (e.g., red)
        cmap.set_bad(color='black')

        # Create the clustermap with horizontal lines
        g = sns.clustermap(grouped_means, cmap=cmap, annot=False,
                        mask=mask, cbar_kws={'label': 'Agreement'},
                        linewidths=0, linecolor='black',
                        col_colors=col_colors, col_cluster=False, row_cluster=False,
                        yticklabels=1)

        # Get the axes object
        ax = g.ax_heatmap

        # # Remove all existing lines
        # ax.grid(False)

        # Add back only horizontal lines
        # ax.set_xticks(np.arange(grouped_means.shape[1]+1)-0.5, minor=True)
        # ax.set_yticks(np.arange(grouped_means.shape[0]+1)-0.5, minor=True)
        # ax.grid(which="minor", color="black", linestyle='-', linewidth=0.5)
        # ax.tick_params(which="minor", bottom=False, left=False)

        # Find where col_colors change
        color_changes = []
        for i in range(1, len(col_colors)):
            if col_colors[i] != col_colors[i-1]:
                color_changes.append(i)

        # Add vertical lines at color change positions
        for pos in color_changes:
            ax.axvline(pos, color='black', linewidth=0.5)

        return g

        # Create a legend for tissues
        # tissue_handles = [plt.Rectangle((0,0),1,1, color=color) for color in tissue_color_map.values()]
        # ax.legend(tissue_handles, tissue_color_map.keys(), title=sub_group_by, 
        #           loc='center left', bbox_to_anchor=(1, 0.5))
        # return fig, ax

    else:
        raise ValueError("Granularity must be 0, 1, or 2.")

    if granularity < 2:
        ax = plt.gca()  # Get current axis for granularity 0 and 1

    ax.set_xlabel(group_by if granularity > 0 else 'Model')
    ax.set_ylabel('Agreement')
    title = 'Average model agreement'
    if granularity == 0:
        title += ''
    elif granularity == 1:
        title += f' by {group_by}'
    elif granularity == 2:
        title += f' by {group_by} and {sub_group_by}'
    ax.set_title(title)

    ax.set_xticklabels(ax.get_xticklabels(), rotation=90, ha='center')

    if granularity < 2:
        ax.legend(title='Models' + ('' if granularity == 0 else ' and Tissues'))

    plt.tight_layout()
    # Return the figure and axis for further editing
    return fig, ax


def plot_model_agreement_categorical(
    adata: AnnData,
    group_by: str,
    sub_group_by: str,
    agreement_cols: list[str],
    granularity: int = 2
) -> tuple[Figure, Axes] | FacetGrid | Figure:
    """
    Plots the relative proportions of categories within specified model columns across varying levels of granularity.

    Parameters
    ------------
    adat 
        An :class:`AnnData`.

    group_by
        key in ``adata.obs`` for the main grouping (e.g., ``'cell_type'``).

    sub_group_by
        key in ``adata.obs`` for the sub-grouping (e.g., ``'tissue'``).

    agreement_cols
        Column names specifying the agreement columns (e.g., ``['categorical_agreement_of_manual_with_model1', 'categorical_agreement_of_manual_with_model2']``). These should be categorical.

    granularity
        level of detail in the plot (``0`` = models only, ``1`` = models within cell types, ``2`` = models within cell types and tissues).

    Returns
    --------
    The plot of agreement, averaged according to the ``granularity`` setting, split by quality of agreement.
    """
    # Verify that the required columns exist
    if not all(col in adata.obs for col in agreement_cols):
        missing_cols = [col for col in agreement_cols if col not in adata.obs]
        raise ValueError(f"Columns {missing_cols} not found in adata.obs.")
    if group_by not in adata.obs:
        raise ValueError(f"Group key '{group_by}' not found in adata.obs.")
    if sub_group_by not in adata.obs:
        raise ValueError(f"Sub-group key '{sub_group_by}' not found in adata.obs.")

    # Ensure that agreement_cols are categorical or convert numeric types to categories
    for col in agreement_cols:
        if not pd.api.types.is_categorical_dtype(adata.obs[col]):
            if pd.api.types.is_numeric_dtype(adata.obs[col]):
                adata.obs[col] = adata.obs[col].astype('category')
            else:
                raise ValueError(f"Column '{col}' must be categorical or convertible to categorical.")

    # Melt the dataframe to get long format
    melted = adata.obs.melt(
        id_vars=[group_by, sub_group_by],
        value_vars=agreement_cols,
        var_name='model_name',
        value_name='agreement'
    )

    # Ensure 'agreement' is categorical and reverse the order of categories
    if not pd.api.types.is_categorical_dtype(melted['agreement']):
        melted['agreement'] = melted['agreement'].astype('category')

    # Reverse the order of 'agreement' categories
    original_categories = melted['agreement'].cat.categories.tolist()
    reversed_categories = original_categories[::-1]
    melted['agreement'] = melted['agreement'].cat.reorder_categories(reversed_categories, ordered=True)

    if granularity == 0:
        # Calculate counts and proportions
        counts = melted.groupby(['model_name', 'agreement']).size().reset_index(name='count')
        total_counts = counts.groupby('model_name')['count'].transform('sum')
        counts['proportion'] = counts['count'] / total_counts

        # Sort models based on total proportion of highest agreement category
        highest_agreement = counts.groupby('model_name')['proportion'].max().reset_index()
        sorted_models = highest_agreement.sort_values('proportion', ascending=False)['model_name']
        counts['model_name'] = pd.Categorical(counts['model_name'], categories=sorted_models, ordered=True)

        # Plot grouped bar chart
        fig, ax = plt.subplots(figsize=(14, 8))
        sns.barplot(
            data=counts,
            x='model_name',
            y='proportion',
            hue='agreement',
            hue_order=reversed_categories,  # Use reversed categories
            ax=ax,
            order=sorted_models
        )

        # Add proportion labels on top of each bar
        for p in ax.patches:
            height = p.get_height()
            if height > 0:
                ax.text(
                    p.get_x() + p.get_width() / 2.,
                    height + 0.01,
                    # f'{height:.2f}',
                    f'{height * 100:.0f}%',
                    ha="center"
                )

        # Rotate x-axis tick labels to vertical
        ax.set_xticklabels(ax.get_xticklabels(), rotation=90)

        ax.set_xlabel('Model')
        ax.set_ylabel('Proportion')
        ax.set_title('Proportion of Agreement Categories by Model')
        ax.set_ylim(0, 1.05)
        ax.legend(title='Agreement Categories', bbox_to_anchor=(1.05, 1), loc='upper left')
        plt.tight_layout()
        return fig, ax

    elif granularity == 1:
        # Calculate counts and proportions
        counts = melted.groupby([group_by, 'model_name', 'agreement']).size().reset_index(name='count')
        total_counts = counts.groupby([group_by, 'model_name'])['count'].transform('sum')
        counts['proportion'] = counts['count'] / total_counts

        # Sort 'group_by' categories based on total proportion
        total_per_group = counts.groupby(group_by)['proportion'].sum().reset_index()
        sorted_groups = total_per_group.sort_values('proportion', ascending=False)[group_by]
        counts[group_by] = pd.Categorical(counts[group_by], categories=sorted_groups, ordered=True)

        # Plot grouped bar chart with model_name as hue
        g = sns.catplot(
            data=counts,
            x=group_by,
            y='proportion',
            hue='agreement',
            hue_order=reversed_categories,  # Use reversed categories
            col='model_name',
            kind='bar',
            height=6,
            aspect=1,
            sharey=True,
            order=sorted_groups
        )

        g.set_axis_labels(group_by, "Proportion")
        g.set_titles("{col_name}")
        g.set(ylim=(0, 1.05))

        # Rotate x-axis tick labels to vertical
        for ax in g.axes.flat:
            ax.set_xticklabels(ax.get_xticklabels(), rotation=90)

        # Add proportion labels on top of each bar
        for ax in g.axes.flatten():
            for p in ax.patches:
                height = p.get_height()
                if height > 0:
                    ax.text(
                        p.get_x() + p.get_width() / 2.,
                        height + 0.01,
                        # f'{height:.2f}',
                        f'{height * 100:.0f}%',
                        ha="center"
                    )

        plt.tight_layout()
        return g

    elif granularity == 2:
        # Calculate counts and proportions
        counts = melted.groupby([group_by, sub_group_by, 'model_name', 'agreement']).size().reset_index(name='count')
        total_counts = counts.groupby([group_by, sub_group_by, 'model_name'])['count'].transform('sum')
        counts['proportion'] = counts['count'] / total_counts

        # Prepare data for heatmap
        pivot_table = counts.pivot_table(
            index=[group_by, sub_group_by],
            columns=['model_name', 'agreement'],
            values='proportion',
            fill_value=0
        )

        # Reverse the order of 'agreement' categories in columns
        pivot_table = pivot_table.reindex(columns=reversed_categories, level=2)

        # Sort index based on total proportion
        pivot_table['Total'] = pivot_table.sum(axis=1)
        pivot_table = pivot_table.sort_values('Total', ascending=False)
        pivot_table = pivot_table.drop(columns='Total')

        # Plot heatmap
        plt.figure(figsize=(14, 8))
        sns.heatmap(
            pivot_table,
            cmap='viridis',
            annot=True,
            fmt=".2f",
            linewidths=0.5
        )
        plt.title(f'Proportion of Agreement Categories by {group_by} and {sub_group_by}')
        plt.tight_layout()
        return plt.gcf()

    else:
        raise ValueError("Granularity must be 0, 1, or 2.")



"""
This module handles LLM querying to annotate cell types in the context of all other cells, based on marker genes alone.
"""

from pandas import DataFrame
from anndata import AnnData

from anndict.utils import enforce_semantic_list
from anndict.llm.llm_call import call_llm
from .base import ai_annotate

def ai_cell_type(gene_list: list[str],
    tissue: str = None
) -> str:
    """
    Returns the cell type based on a list of marker genes as determined by LLM.

    Parameters
    -------------
    gene_list
        The list of genes to be described.

    tissue
        The tissue of origin to provide context for the LLM.

    Returns
    --------
    The cell type label generated by an LLM.
    """
    if not gene_list:
        raise ValueError("Empty gene list passed to ai_cell_type")

    #enforce that labels are semantic
    enforce_semantic_list(gene_list)

    # Prepare the prompt
    if len(gene_list) == 1:
        base_prompt = f"In a few words and without restating any part of the question, describe the single most likely cell type represented by the marker gene: {gene_list[0]}"
    else:
        genes_str = "    ".join(gene_list)
        base_prompt = f"In a few words and without restating any part of the question, describe the single most likely cell type represented by the marker genes: {genes_str}"

    # Add tissue information if provided
    if tissue:
        base_prompt += f" Consider that these cells are from {tissue} tissue."

    # Prepare the messages for the Chat Completions API
    messages = [
        {"role": "system", "content": "You are a terse molecular biologist."},
        {"role": "user", "content": base_prompt}
    ]

    # Call the LLM using the call_llm function
    annotation = call_llm(
        messages=messages,
        max_tokens=100,
        temperature=0
    )

    return annotation

def ai_annotate_cell_type(adata: AnnData,
    groupby: str,
    n_top_genes: int,
    new_label_column: str = 'ai_cell_type',
    tissue_of_origin_col: str = None
) -> DataFrame:
    """
    Annotate cell types based on the top marker genes for each cluster.

    This function performs differential expression analysis to identify marker genes for each cluster
    and applies ai_annotate_cell_type to determine the cell type for each cluster based on the top 
    marker genes. The results are added to the AnnData object and returned as a DataFrame.

    Parameters
    ----------
    adata
        Column in adata.obs to group by for differential expression analysis.
    groupby
        Column in adata.obs to group by for differential expression analysis.
    n_top_genes
        The number of top marker genes to consider for each cluster.
    new_label_column
        The name of the new column in adata.obs where the cell type annotations will be stored.
        Default: 'ai_cell_type'
    tissue_of_origin_col
        Missing description for this parameter.
        Default: None

    Returns
    -------
    A pd.DataFrame with a column for the top marker genes for each cluster.
    """
    return ai_annotate(func=ai_cell_type, adata=adata, groupby=groupby, n_top_genes=n_top_genes, new_label_column=new_label_column, tissue_of_origin_col=tissue_of_origin_col)



"""
This subpackage contains functions to annotate and assess cell and gene annotation.
"""

#gene annotation functions
from .genes import (
    ai_make_cell_type_gene_list
)

#cell annotation functions
from .cells import (
    # De Novo
    ai_annotate,
    ai_annotate_by_comparison,
    ai_cell_type,
    ai_annotate_cell_type,
    ai_cell_types_by_comparison,
    ai_annotate_cell_type_by_comparison,
    ai_annotate_cell_type_by_comparison_adata_dict,
    ai_annotate_cell_sub_type,
    ai_annotate_biological_process,
    cell_type_marker_gene_score,
    ai_determine_leiden_resolution,

    # Label Transfer
    harmony_label_transfer,
    stable_label, stable_label_adata,

    # Benchmarking
    create_label_df,
    kappa_adata, krippendorff_alpha_adata,
    ai_compare_cell_type_labels_pairwise
)

__all__ = [
    # Genes
    #--------
    # genes.py
    "ai_make_cell_type_gene_list",

    # Cells
    #--------
    # De Novo
    "ai_annotate",
    "ai_annotate_by_comparison",
    "ai_cell_type",
    "ai_annotate_cell_type",
    "ai_cell_types_by_comparison",
    "ai_annotate_cell_type_by_comparison",
    "ai_annotate_cell_type_by_comparison_adata_dict",
    "ai_annotate_cell_sub_type",
    "ai_annotate_biological_process",
    "cell_type_marker_gene_score",
    "ai_determine_leiden_resolution",

    # Label Transfer
    "harmony_label_transfer",
    "stable_label",
    "stable_label_adata",

    # Benchmarking
    "create_label_df",
    "kappa_adata",
    "krippendorff_alpha_adata",
    "ai_compare_cell_type_labels_pairwise"
]


"""
This module contains utility functions for label comparison.
"""

import pandas as pd

def create_label_df(adata, cols1, cols2):
    """
    Create a DataFrame of unique label combinations from the specified columns in cols1 and cols2,
    only including combinations that exist in adata.obs.
    Parameters:
    adata: AnnData object containing the data.
    cols1: List of columns to compare against cols2.
    cols2: List of columns to compare with cols1.
    Returns:
    pd.DataFrame: DataFrame containing unique combinations of the specified columns.
    """
    # Combine all columns
    all_cols = cols1 + cols2

    # Get unique combinations that exist in adata.obs
    unique_combinations = adata.obs[all_cols].drop_duplicates()

    # Melt the DataFrame to get all combinations in two columns
    melted_df = pd.melt(unique_combinations,
                        id_vars=cols1,
                        value_vars=cols2,
                        var_name='col2_name',
                        value_name='col2')

    # Melt again to get col1 in a single column
    result_df = pd.melt(melted_df,
                        id_vars=['col2_name', 'col2'],
                        value_vars=cols1,
                        var_name='col1_name',
                        value_name='col1')

    # Keep only the relevant columns and drop duplicates
    result_df = result_df[['col1', 'col2']].drop_duplicates()

    return result_df


"""
The base class for managing LLM provider initialization and rate limiting.

The module defines:
- ``LLMProviderConfig``: Configuration dataclass for LLM providers
- ``BaseLLMInitializer``: Abstract base class for provider initialization logic 
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Type
from langchain_core.rate_limiters import InMemoryRateLimiter


@dataclass
class LLMProviderConfig:
    """Configuration for LLM provider"""

    class_name: str
    module_path: str
    init_class: Type["BaseLLMInitializer"]
    requests_per_minute: float = 40
    check_every_n_seconds: float = 0.1


class BaseLLMInitializer(ABC):
    """Base class for LLM initialization logic"""

    def __init__(self, config: LLMProviderConfig):
        self.config = config

    def create_rate_limiter(
        self, constructor_args: dict[str, Any]
    ) -> InMemoryRateLimiter:
        """Create a rate limiter with given or default parameters"""
        requests_per_minute = float(
            constructor_args.pop("requests_per_minute", self.config.requests_per_minute)
        )
        requests_per_second = requests_per_minute / 60
        check_every_n_seconds = float(
            constructor_args.pop(
                "check_every_n_seconds", self.config.check_every_n_seconds
            )
        )
        max_bucket_size = float(
            constructor_args.pop("max_bucket_size", requests_per_minute)
        )

        return InMemoryRateLimiter(
            requests_per_second=requests_per_second,
            check_every_n_seconds=check_every_n_seconds,
            max_bucket_size=max_bucket_size,
        )

    @abstractmethod
    def initialize(
        self, constructor_args: dict[str, Any], **kwargs
    ) -> tuple[dict[str, Any], dict[str, Any]]:
        """Initialize the LLM provider with given arguments"""


"""
This subpackage contains functions to annotate cells de novo (from scratch), based on marker genes.
"""

#the base ai_annotate functions
from .base import (
    ai_annotate,
    ai_annotate_by_comparison
)

#cell annotation methods--by gene list
from .annotate_cell_type import (
    ai_cell_type,
    ai_annotate_cell_type
)

#cell type annotation methods--by gene list (in the context of other gene lists)
from .annotate_cell_type_by_comparison import (
    ai_cell_types_by_comparison,
    ai_annotate_cell_type_by_comparison,

)

from .annotate_cell_subtype import (
    ai_annotate_cell_type_by_comparison_adata_dict,
    ai_annotate_cell_sub_type,
)

#annotate groups of cells with biological process (and not cell type)
from .annotate_biological_process import (
    ai_annotate_biological_process,
)

#automatically calculate gene module scores for each cell type
from .cell_type_score import (
    cell_type_marker_gene_score,
)

#automatically determine cluster resolution
from .automated_clustering import (
    ai_determine_leiden_resolution,
)

__all__ = [
    # base.py
    "ai_annotate",
    "ai_annotate_by_comparison",

    # annotate_cell_type.py
    "ai_cell_type",
    "ai_annotate_cell_type",

    # annotate_cell_type_by_comparison.py
    "ai_cell_types_by_comparison",
    "ai_annotate_cell_type_by_comparison",
    

    # annotate_cell_subtype.py
    "ai_annotate_cell_type_by_comparison_adata_dict",
    "ai_annotate_cell_sub_type",

    # annotate_biological_process.py
    "ai_annotate_biological_process",

    # cell_type_score.py
    "cell_type_marker_gene_score",

    # automated_clustering.py
    "ai_determine_leiden_resolution"
]



"""
The module contains functions to annotate cell subtype by comparison.
Specifically, this means asking an LLM to annotate a subtype by comparing its marker genes
to the marker genes of other subtypes of that cell type).
"""

from .base import ai_annotate_by_comparison


"""
AnnDictionary utility functions.
"""

import os
import re

import pandas as pd
import seaborn as sns

import matplotlib
import matplotlib.pyplot as plt

from IPython.display import HTML, display
from anndata import AnnData


def enforce_semantic_list(
    lst: list[str]
) -> None:
    """
    This function runs a number of checks to make sure that 
    the input is a semantic list, and not i.e. integers cast as strings.

    Parameters
    -----------
    lst
        List of strings to be checked.

    Returns
    --------
    None

    Raises
    ------
    ValueError
        If the input list contains any of: NaN, numeric, or numeric cast as string.

    Examples
    --------
    .. code-block:: python

        enforce_semantic_list(['gene1', 'gene2', 'gene3'])

    Notes
    -----
    This function is useful for ensuring that the input list 
    contains semantic labels (i.e. gene symbols or cell types) and 
    not integer labels for LLM interpretation. It checks if all items are 
    strings and if any item can be converted to float.
    """
    error_message = "input list appears to contain any of: NaN, numeric, or \
        numeric cast as string. Please ensure you are passing semantic labels \
        (i.e. gene symbols or cell types) and not integer labels for AI \
        interpretation. Make sure adata.var.index and adata.obs.index are \
        not integers or integers cast as strings."

    def get_context(lst, index):
        before = lst[index - 1] if index > 0 else None
        after = lst[index + 1] if index < len(lst) - 1 else None
        return before, after

    # Check if all items are strings
    for index, item in enumerate(lst):
        if not isinstance(item, str):
            before, after = get_context(lst, index)
            raise ValueError(
                f"{error_message} Item at index {index} is not a \
                string: {item}. Context: Before: {before}, After: {after}"
            )

    # Check if any item can be converted to float
    for index, item in enumerate(lst):
        try:
            float(item)
        except ValueError:
            pass
        else:
            before, after = get_context(lst, index)
            raise ValueError(
                f"{error_message} Item at index {index} can be cast to \
                a number: {item}. Context: Before: {before}, After: {after}"
            )

    return True


def make_names(
    names: list[str]
) -> list[str]:
    """
    Convert a list of names into valid and unique Python identifiers.

    Parameters
    ----------
    names
        List of names to be transformed.

    Returns
    -------
        Valid and unique Python identifiers.
    """
    # Equivalent of R's make.names() function in Python
    valid_names = []
    seen = {}
    for name in names:
        # Replace invalid characters with underscores
        clean_name = re.sub(r"[^0-9a-zA-Z_]", "_", name)
        if clean_name in seen:
            seen[clean_name] += 1
            clean_name = f"{clean_name}.{seen[clean_name]}"
        else:
            seen[clean_name] = 0
        valid_names.append(clean_name)
    return valid_names


def normalize_string(
    s: str
) -> str:
    """
    Removes non-alphanumeric characters and converts to lowercase.

    Parameters
    ----------
    s
        String to be normalized.

    Returns
    -------
        Normalized string.

    Examples
    --------
    .. code-block:: python

        normalize_string("Hello, World!")
        > "hello world"
    """
    return re.sub(r"[^\w\s]", "", s.lower())


def normalize_label(
    label: str
) -> str:
    """
    Calls normalize-string and handles NaN values.

    Parameters
    ----------
    label
        Label to be normalized.

    Returns
    -------
        Normalized string.

    Examples
    --------
    .. code-block:: python

        normalize_label("Hello, World!")
        > "hello world"

    .. code-block:: python

        normalize_label(np.nan)
        > "missing"
    """
    if pd.isna(label):  # Handle NaN values
        return "missing"
    return normalize_string(label.strip())


def create_color_map(adata: AnnData,
    keys: list[str]
) -> dict:
    """
    Creates a unified color map for given keys from an :class:`AnnData` object, differentiating
    between continuous and categorical data.

    Parameters
    ----------
    adata
        An :class`AnnData`.
    keys
        keys for which the color map is required.

    Returns
    -------
    A color map linking unique values or ranges from the specified keys to colors.
    """
    color_map = {}
    for key in keys:
        if pd.api.types.is_numeric_dtype(adata.obs[key]):
            # Create a continuous colormap
            min_val, max_val = adata.obs[key].min(), adata.obs[key].max()
            norm = plt.Normalize(min_val, max_val)
            scalar_map = plt.cm.ScalarMappable(norm=norm, cmap="viridis")
            # Store the scalar map directly
            color_map[key] = scalar_map
        else:
            # Handle as categorical
            unique_values = pd.unique(adata.obs[key])
            color_palette = sns.color_palette("husl", n_colors=len(unique_values))
            color_palette_hex = [
                matplotlib.colors.rgb2hex(color) for color in color_palette
            ]
            color_map[key] = dict(zip(unique_values, color_palette_hex))

    return color_map


def get_slurm_cores():
    """
    Returns the total number of CPU cores allocated to a Slurm job based on environment variables.

    Returns
    -------
    The total number of CPU cores allocated to the Slurm job.

    Examples
    --------
    .. code-block:: python

        get_slurm_cores()
        > 40

    Notes
    -----
    Checks the environment variables ``SLURM_CPUS_PER_TASK`` and ``SLURM_NTASKS`` to
    calculate the total number of CPU cores allocated to the job by multiplying their values.

    For environment variables that are not set, the function defaults to 1 CPU core.
    """
    # Get the number of CPUs per task (default to 1 if not set)
    cpus_per_task = int(os.getenv("SLURM_CPUS_PER_TASK", "1"))

    # Get the number of tasks (default to 1 if not set)
    ntasks = int(os.getenv("SLURM_NTASKS", "1"))

    # Calculate total cores
    total_cores = cpus_per_task * ntasks

    return total_cores


def summarize_metadata(
    adata: AnnData,
    cols: str | list[str]
) -> dict:
    """
    Generate a summary for specified metadata column(s) in ``adata``.

    Parameters
    -----------
    adata
        The anndata object containing the data.

    columns
        List of columns in ``adata.obs``. Use '*' to specify joint frequencies of multiple columns.

    Returns
    --------
    A :class:`dict` with keys as column descriptions and values as a :class:`DataFrame` of counts.

    Notes
    -----
    Use '*' to specify joint frequencies of multiple columns.

    Examples
    ---------

    **Case 1: calculate frequency of donor and tissue columns**

    .. code-block:: python

        import anndict as adt
        adt.wrappers.summarize_metadata(adata, ['donor', 'tissue'])

    **Case 2: calculate joint frequency of donor x tissue**

    .. code-block:: python

        import anndict as adt
        adt.wrappers.summarize_metadata(adata, ['donor*tissue'])

    See Also
    ---------
    :func:`get_adata_columns` : To use string matching to 
        retrieve specific column names from ``adata``.
    :func:`display_html_summary` : To print the results of 
        :func:`summarize_metadata` as an html table.
    :func:`anndict.wrappers.anndictionary_.summarize_metadata_adata_dict` :  
    :func:`anndict.wrappers.anndictionary_.display_html_summary_adata_dict` : 
    """
    results = {}

    for col in cols:
        if "*" in col:
            # Handle joint frequencies
            sub_cols = col.split("*")
            combined_data = adata.obs[sub_cols]

            # Convert categorical columns to string and replace NaN with a placeholder
            for sub_col in sub_cols:
                if isinstance(combined_data[sub_col].dtype, pd.CategoricalDtype):
                    combined_data[sub_col] = combined_data[sub_col].astype(str)

            # Replace NaN with a placeholder to include them in groupby
            combined_data = combined_data.fillna("NaN")

            joint_freq = combined_data.groupby(sub_cols).size().unstack(fill_value=0)
            joint_freq = (
                combined_data.groupby(sub_cols, observed=True)
                .size()
                .unstack(fill_value=0)
            )
            results[col.replace("*", " x ")] = joint_freq
        else:
            # Calculate frequency for a single column
            freq = adata.obs[col].value_counts(dropna=False).to_frame("count")
            results[col] = freq

    return results


def display_html_summary(
    summary_dict,
    return_html=False
) -> str:
    """
    Display separate HTML tables for each metadata category in the summary dictionary,
    arranging up to three tables in a row before starting a new line.

    Parameters
    ----------
    summary_dict
        The :class:`dict` containing frequency data for metadata columns.

    Returns
    -------
    If ``return_html`` is ``True``, returns the HTML string. Otherwise, returns ``None``.

    Notes
    -----
    Displays the results of :func:`summarize_metadata` as an HTML table.

    Examples
    --------

    .. code-block:: python

        import anndict as adt
        summary_dict = adt.wrappers.summarize_metadata(adata, ['donor', 'tissue', 'donor*tissue'])
        adt.wrappers.display_html_summary(summary_dict)

    See Also
    --------
    :func:`summarize_metadata` : To generate the summary data.
    """
    html = '<div style="display: flex; flex-wrap: wrap;">'
    table_count = 0

    for category, data in summary_dict.items():
        if table_count % 3 == 0 and table_count != 0:
            html += '<div style="flex-basis: 100%; height: 20px;"></div>'

        table_html = f'<div style="flex: 1; padding: 10px;"><h3>{category}</h3>'
        # Start the table and add a header row
        table_html += (
            '<table border="1"><tr><th></th>'  # Empty header for the row labels
        )
        table_html += (
            "".join(f"<th>{col}</th>" for col in data.columns) + "</tr>"
        )  # Column headers

        for index, row in data.iterrows():
            # Include row labels as the first column and the rest of the data in subsequent columns
            table_html += (
                f"<tr><td>{index}</td>"
                + "".join(f"<td>{val}</td>" for val in row)
                + "</tr>"
            )

        table_html += "</table></div>"
        html += table_html
        table_count += 1

    html += "</div>"
    display(HTML(html))

    if return_html:
        return html



"""
Plot confusion matrices. Inlcudes the option to rearrange columns and rows to make the matrix appear diagonal, which can facilitate interpretation.
"""

import numpy as np
import pandas as pd

import seaborn as sns
import matplotlib.pyplot as plt

from scipy.optimize import linear_sum_assignment
from sklearn.metrics import confusion_matrix
from sklearn.preprocessing import LabelEncoder
from seaborn.matrix import ClusterGrid
from anndata import AnnData


from anndict.utils import create_color_map

def plot_confusion_matrix_from_adata(
    adata: AnnData,
    true_label_key: str,
    predicted_label_key: str,
    title: str = 'Confusion Matrix',
    row_color_keys: str | None = None,
    col_color_keys: str | None = None,
    figsize: tuple[int,int] | None = None,
    diagonalize: bool = False,
    true_ticklabels: list[str] | None = None,
    predicted_ticklabels: list[str] | None = None,
    annot: bool | None = None
) -> ClusterGrid:
    """
    Plots a confusion matrix from an :class:`AnnData` object, with optional row and column colors. 
    Set ``diagonalize=True`` to automatically arrange the column and rows to make the confusion matrix appear as diagonal as possible.
    Wraps :func:`plot_confusion_matrix`.
    
    Parameters
    ------------
    adata 
        An :class:`AnnData`.

    true_label_key
        key to access the true class labels in ``adata.obs``.

    predicted_label_key
        key to access the predicted class labels in ``adata.obs``.

    title
        title of the plot.

    row_color_key
        key for row colors in ``adata.obs``.

    col_color_key
        key for column colors in ``adata.obs``.

    Returns
    --------
    The confusion matrix plot object.

    See Also
    ---------
    :func:`plot_confusion_matrix` : the main plotting function used.
    """

    # Check and convert row_color_key and col_color_key to lists if they are not None
    if row_color_keys is not None and not isinstance(row_color_keys, list):
        row_color_keys = [row_color_keys]

    if col_color_keys is not None and not isinstance(col_color_keys, list):
        col_color_keys = [col_color_keys]

    # Get unique labels
    true_labels = adata.obs[true_label_key].astype(str)
    predicted_labels = adata.obs[predicted_label_key].astype(str)

    combined_labels = pd.concat([true_labels, predicted_labels])
    label_encoder = LabelEncoder()
    label_encoder.fit(combined_labels)

    #Encode labels
    true_labels_encoded = label_encoder.transform(true_labels)
    predicted_labels_encoded = label_encoder.transform(predicted_labels)

    # Create label-to-color dictionary for mapping
    true_label_color_dict = None
    if row_color_keys:
        true_label_subset = adata.obs[[true_label_key] + row_color_keys].drop_duplicates().set_index(true_label_key)
        true_label_color_dict = {label: {key: row[key] for key in row_color_keys}
                        for label, row in true_label_subset.iterrows()
                        }

    predicted_label_color_dict = None
    if col_color_keys:
        predicted_label_subset = adata.obs[[predicted_label_key] + col_color_keys].drop_duplicates().set_index(predicted_label_key)
        predicted_label_color_dict = {label: {key: col[key] for key in col_color_keys}
                        for label, col in predicted_label_subset.iterrows()
                        }

    # Compute the row and column colors
    # Get unified color mapping
    keys = list(set(row_color_keys or []).union(col_color_keys or []))
    color_map = create_color_map(adata, keys)

    # Call the main plot function
    return plot_confusion_matrix(true_labels_encoded, predicted_labels_encoded, label_encoder, color_map, title,
                          row_color_keys=row_color_keys, col_color_keys=col_color_keys,
                          true_label_color_dict=true_label_color_dict, predicted_label_color_dict=predicted_label_color_dict,
                          figsize=figsize, diagonalize=diagonalize, true_ticklabels=true_ticklabels,
                          predicted_ticklabels=predicted_ticklabels, annot=annot)


def plot_confusion_matrix(
    true_labels_encoded: np.ndarray,
    predicted_labels_encoded: np.ndarray,
    label_encoder: LabelEncoder,
    color_map: dict[str, dict[str, str]],
    title: str = 'Confusion Matrix', 
    row_color_keys: list[str] | None = None,
    col_color_keys: list[str] | None = None,
    true_label_color_dict: dict[str, dict[str, str]] | None = None,
    predicted_label_color_dict: dict[str, dict[str, str]] | None = None,
    figsize: tuple[float, float] | None = None,
    diagonalize: bool = False,
    true_ticklabels: bool | list[str] | None = None,
    predicted_ticklabels: bool | list[str] | None = None,
    annot: bool | None = None
) -> ClusterGrid:
    """
    Plot a confusion matrix using cluster mapping with optional color annotations.
    
    This function computes a confusion matrix based on encoded true and predicted labels
    and optionally reorders the matrix to maximize diagonal alignment. It supports
    row and column coloring based on custom dictionaries and color maps.
    
    Parameters
    ----------
    true_labels_encoded
        Encoded true labels.

    predicted_labels_encoded
        Encoded predicted labels.

    label_encoder
        A fitted ``sklearn`` ``LabelEncoder`` that can decode the labels.

    color_map
        A nested dictionary specifying how labels should be colored.
        The first-level keys match possible color keys (e.g., ``'row'``, ``'col'``),
        and each maps to a dict from label attributes to color codes.

    title
        The title of the plot.

    row_color_keys
        Keys to extract color information for the rows from ``color_map``.

    col_color_keys
        Keys to extract color information for the columns from ``color_map``.

    true_label_color_dict
        A nested :class:`dict` specifying row-label-specific color mappings.

    predicted_label_color_dict
        A nested :class:`dict` specifying column-label-specific color mappings.

    true_labels
        Original (not encoded) true labels, if needed.

    predicted_labels
        Original (not encoded) predicted labels, if needed.

    figsize
        Figure size passed to seaborn's :func:`clustermap`.

    diagonalize
        If ``True``, reorders the confusion matrix to maximize its diagonal alignment.

    true_ticklabels
        Labels to display along the rows of the confusion matrix.
        If ``True``, display all; if ``False``, display none.

    predicted_ticklabels
        Labels to display along the columns of the confusion matrix.
        If ``True``, display all; if ``False``, display none.

    annot
        Whether to annotate each cell with its numeric value.

    Returns
    --------
    A seaborn :class:`ClusterGrid` object displaying the plotted confusion matrix.

    Notes
    ------
    - If the number of true or predicted labels exceeds 40, tick labels and annotations are disabled by default for better visibility.
    - The matrix is normalized by the number of samples in each true class.
    - If `diagonalize` is True, reorder the columns and row to diagonalize the matrix to the extent possible.

    Examples
    ---------
    >>> # Assume true_encoded, pred_encoded, and le are given, along with a color_map dict
    >>> g = plot_confusion_matrix(
    ...     true_labels_encoded=true_encoded,
    ...     predicted_labels_encoded=pred_encoded,
    ...     label_encoder=le,
    ...     color_map={'some_color_key': {'some_label_value': '#ff0000'}}
    ... )
    """

    labels_true = np.unique(true_labels_encoded)
    labels_pred = np.unique(predicted_labels_encoded)

    # Compute the confusion matrix
    cm = confusion_matrix(true_labels_encoded, predicted_labels_encoded, labels=np.arange(len(label_encoder.classes_)))

    # Normalize the confusion matrix by row (i.e., by the number of samples in each class)
    cm_normalized = cm.astype('float') / cm.sum(axis=1, keepdims=True)
    cm_normalized = pd.DataFrame(cm_normalized[np.ix_(labels_true, labels_pred)], 
                                 index=label_encoder.inverse_transform(labels_true), 
                                 columns=label_encoder.inverse_transform(labels_pred))

    if diagonalize:
        # Sorting the confusion matrix to make it as diagonal as possible
        cost_matrix = -cm_normalized.values  # We need to minimize the cost, hence the negative sign
        row_ind, col_ind = linear_sum_assignment(cost_matrix)

        # Concatenate the optimal indices with the non-optimal ones
        row_ind = np.concatenate((row_ind, np.setdiff1d(np.arange(cm_normalized.shape[0]), row_ind)))
        col_ind = np.concatenate((col_ind, np.setdiff1d(np.arange(cm_normalized.shape[1]), col_ind)))

        cm_normalized = cm_normalized.iloc[row_ind, col_ind]
        labels_true_sorted = label_encoder.inverse_transform(labels_true)[row_ind]
        labels_pred_sorted = label_encoder.inverse_transform(labels_pred)[col_ind]
    else:
        labels_true_sorted = label_encoder.inverse_transform(labels_true)
        labels_pred_sorted = label_encoder.inverse_transform(labels_pred)

    def map_labels_to_colors(labels, label_color_dict, color_map):
        color_list = []
        for label in labels:
            color_dict = label_color_dict.get(label, {})
            colors = [color_map.get(key).get(color_dict.get(key, None), '#FFFFFF') for key in color_map]
            color_list.append(colors)
        return color_list

    row_colors = None
    if row_color_keys:
        row_colors = map_labels_to_colors(labels_true_sorted, true_label_color_dict, color_map)
        row_colors = pd.DataFrame(row_colors, index=labels_true_sorted)

    col_colors = None
    if col_color_keys:
        col_colors = map_labels_to_colors(labels_pred_sorted, predicted_label_color_dict, color_map)
        col_colors = pd.DataFrame(col_colors, index=labels_pred_sorted)

    xticklabels = predicted_ticklabels if predicted_ticklabels is not None else (True if len(labels_pred) <= 40 else False)
    yticklabels = true_ticklabels if true_ticklabels is not None else (True if len(labels_true) <= 40 else False)
    annot = annot if annot is not None else (True if len(labels_true) <= 40 and len(labels_pred) <= 40 else False)


    g = sns.clustermap(cm_normalized, annot=annot, fmt=".2f", cmap="Blues",
                       row_colors=row_colors, col_colors=col_colors,
                       xticklabels=xticklabels, yticklabels=yticklabels,
                       row_cluster=False, col_cluster=False, figsize=figsize)

    g.ax_heatmap.set_title(title, y=1.05)
    g.ax_heatmap.set_ylabel('True label')
    g.ax_heatmap.set_xlabel('Predicted label')
    plt.show()

    return g
